# take nth order derivatives
from sympy import symbols,Function,tanh
from sympy import diff as D
import numpy as np
#(xyzk,KDTree=None,patch_points=None,b_coef=None,h=None,
# q=None,sol_mat=None,patch_x=None,patch_y=None,patch_z=None):
def outter_parentheses(string):
    open_p = [i for i, ch in enumerate(string) if ch == "("]
    close_p = [i for i, ch in enumerate(string) if ch == ")"]
    pairs = []
    while len(close_p) > 0:
        j = close_p[0]
        i = None
        jj = j
        while i == None:
            i = string[:jj].rfind("(")
            if i in open_p:
                break
            else:
                jj = i
                i = None
        if i == open_p[0]:
            pairs.append((i,j+1))
        close_p.remove(j)
        open_p.remove(i)
    return pairs

def global_d(order):
    x,y,z    = symbols('x y z')
    xi,yi,zi = symbols('xi yi zi')
    A,B1,B2,B3,C1,C2,C3,C4 = symbols('A B1 B2 B3 C1 C2 C3 C4')
    r  = ((x-xi)**2+(y-yi)**2+(z-zi)**2)**(1/2)
    rr = ((x-xi)**2+(y-yi)**2+(z-zi)**2)
    dfix = D(r**3,x)
    dfiy = D(r**3,y)
    dfiz = D(r**3,z)
    fi = -tanh(A*r**3+B1*dfix+B2*dfiy+B3*dfiz+C1*x+C2*y+C3*z+C4)
    fi_inner = A*r**3+B1*dfix+B2*dfiy+B3*dfiz+C1*x+C2*y+C3*z+C4
    fi_inner = str(fi_inner)
    fi_inner = fi_inner.replace(str(rr),'rr')
    fs = str(fi_inner)
    vars = [x,y,z]
    del_orders_f = [[fi]]
    del_notation = [['f']]
    for i in range(order):
        tmp_f = []
        tmp_n = []
        for idx,p in enumerate(del_orders_f[-1]):
            for jdx,v in enumerate(vars):
                tmp_f.append(D(p,v))
                if i == 0:
                   tmp_n.append('d'+del_notation[-1][idx]+str(v))
                else:
                   tmp_n.append(del_notation[-1][idx]+str(v))
        del_orders_f.append(tmp_f)
        del_notation.append(tmp_n)
    key = {str(rr) :'rr',
           'tanh'  :'np.tanh',
           fi_inner:'f'}
    var = {'A' :'sol_mat[0,:,0]',
           'B1':'sol_mat[0,:,1]',
           'B2':'sol_mat[0,:,2]',
           'B3':'sol_mat[0,:,3]',
           'C1':'b_coef[0,:,1]',
           'C2':'b_coef[0,:,2]',
           'C3':'b_coef[0,:,3]',
           'C4':'b_coef[0,:,0]',
           'xi':'patch_points[0,:,0]',
           'yi':'patch_points[0,:,1]',
           'zi':'patch_points[0,:,2]',
           'x' :'pt_3[0,:,0]',
           'y' :'pt_3[0,:,1]',
           'z' :'pt_3[0,:,2]',}
    for i in range(len(del_orders_f)):
        for j in range(len(del_orders_f[i])):
            del_orders_f[i][j] = str(del_orders_f[i][j])
            for item in key.keys():
                if item in del_orders_f[i][j]:
                    del_orders_f[i][j] = del_orders_f[i][j].replace(item,key[item])
            for item in var.keys():
                if item in del_orders_f[i][j]:
                    del_orders_f[i][j] = del_orders_f[i][j].replace(item,var[item])
                if item in fs:
                    fs = fs.replace(item,var[item])
    return del_orders_f,del_notation,fs

def construct_global(order,dim=3):
    funcs,nots,fs = global_d(order)
    functions = []
    func_strings = []
    for n in range(len(funcs)):
        d_func = 'def d_{}(xyz,'.format(n)+\
                           'patch_points=None,b_coef=None,'+\
                           'sol_mat=None):\n'
        d_func += '    x,y,z = xyz\n'
        d_func += '    pt = np.ones((1,b_coef.shape[-1]))\n'
        d_func += '    pt[:,1] *= x\n'
        d_func += '    pt[:,2] *= y\n'
        d_func += '    pt[:,3] *= z\n'
        d_func += '    pt_3 = np.array([[[x,y,z]]])\n'
        d_func += '    rr = np.sum(np.square(pt_3-patch_points[:,:]),axis=2)\n'
        if n == 0:
            d_func += '    {} = {}\n'.format(nots[0][0],fs)
            d_func += '    {} = np.sum({},axis=1)\n'.format(nots[0][0],nots[0][0])
        elif n > 0:
            funcs[0][0] = fs
        for i in range(n+1):
            for j in range(len(funcs[i])):
                tmp_func = funcs[i][j]
                d_func += '    {} = {}\n'.format(nots[i][j],funcs[i][j])
                if i == 0 and n > 0:
                    d_func += '    {} = np.sum({},axis=1)\n'.format(nots[i][j],funcs[i][j])
        d_func += '    res = np.zeros({})\n'.format(len(funcs[n]))
        if n > 0:
            d_func += '    {} = -np.tanh({})\n'.format(nots[0][0],nots[0][0])
        for j in range(len(funcs[n])):
            d_func += '    res_{} = {}\n'.format(nots[n][j],nots[n][j])
            d_func += '    res_{} = np.sum(res_{})\n'.format(nots[n][j],nots[n][j])
            d_func += '    res[{}] = res_{}\n'.format(j,nots[n][j])
        if n > 1:
            d_func += '    res = res.reshape({}'.format(dim)
            for c in range(n-1):
                d_func += ',{}'.format(dim)
            d_func += ')\n'
        d_func += '    return res\n'
        #d_func += 'functions.append(d_{})'.format(n)
        exec(d_func)
        func_strings.append(d_func)
    return functions,func_strings

def d(order):
    # Symbolically express nth order derivatives
    x,y,z    = symbols('x y z')
    xi,yi,zi = symbols('xi yi zi')
    xc,yc,zc = symbols('xc yc zc')
    A,B1,B2,B3,C1,C2,C3,C4 = symbols('A B1 B2 B3 C1 C2 C3 C4')
    h,q = symbols('h q')
    r  = ((x-xi)**2+(y-yi)**2+(z-zi)**2)**(1/2)
    rr = ((x-xi)**2+(y-yi)**2+(z-zi)**2)
    rc = ((x-xc)**2+(y-yc)**2+(z-zc)**2)**(1/2)
    rrc = ((x-xc)**2+(y-yc)**2+(z-zc)**2)
    dfix = D(r**3,x)
    dfiy = D(r**3,y)
    dfiz = D(r**3,z)
    fi = -tanh(A*r**3+B1*dfix+B2*dfiy+B3*dfiz+C1*x+C2*y+C3*z+C4)
    fi_inner = A*r**3+B1*dfix+B2*dfiy+B3*dfiz+C1*x+C2*y+C3*z+C4
    fi_inner = str(fi_inner)
    fi_inner = fi_inner.replace(str(rr),'rr')
    fs = str(fi_inner)
    wi = (1-tanh(rc))/(h+rc)**q
    Fi = Function('Fi')(x,y,z)
    Wi = Function('Wi')(x,y,z)
    Ki  = Function('K')(x,y,z)
    F  = (Fi*Wi)/Ki
    vars = [x,y,z]
    if order == 0:
        del_orders_fi = [[fi]]
        del_orders_wi = [[wi]]
        del_orders_F  = [[F]]
        del_orders_Fi = [[Fi]]
        del_orders_Wi = [[Wi]]
        del_orders_Ki = [[Ki]]
        notation_f    = [['f']]
        notation_w    = [['w']]
        notation_k    = [['k']]
    else:
        del_orders_fi = []
        del_orders_wi = []
        del_orders_F  = []
        del_orders_Fi = []
        del_orders_Wi = []
        del_orders_Ki = []
        notation_f    = []
        notation_w    = []
        notation_k    = []
    for i in range(order):
        tmp_del_order_fi = []
        tmp_del_order_wi = []
        tmp_del_order_F  = []
        tmp_del_order_Fi = []
        tmp_del_order_Wi = []
        tmp_del_order_Ki = []
        tmp_notation_f   = []
        tmp_notation_w   = []
        tmp_notation_k   = []
        if i == 0:
            for j in range(len(vars)):
                tmp_del_order_fi.append(D(fi,vars[j]))
                tmp_del_order_wi.append(D(wi,vars[j]))
                tmp_del_order_F.append(D(F,vars[j]))
                tmp_del_order_Fi.append(D(Fi,vars[j]))
                tmp_del_order_Wi.append(D(Wi,vars[j]))
                tmp_del_order_Ki.append(D(Ki,vars[j]))
                tmp_notation_f.append('df'+str(vars[j]))
                tmp_notation_w.append('dw'+str(vars[j]))
                tmp_notation_k.append('dk'+str(vars[j]))
        else:
            for p in range(len(del_orders_F[-1])):
                for j in range(len(vars)):
                    tmp_del_order_fi.append(D(del_orders_fi[-1][p],vars[j]))
                    tmp_del_order_wi.append(D(del_orders_wi[-1][p],vars[j]))
                    tmp_del_order_F.append(D(del_orders_F[-1][p],vars[j]))
                    tmp_del_order_Fi.append(D(del_orders_Fi[-1][p],vars[j]))
                    tmp_del_order_Wi.append(D(del_orders_Wi[-1][p],vars[j]))
                    tmp_del_order_Ki.append(D(del_orders_Ki[-1][p],vars[j]))
                    tmp_notation_f.append(notation_f[-1][p]+str(vars[j]))
                    tmp_notation_w.append(notation_w[-1][p]+str(vars[j]))
                    tmp_notation_k.append(notation_k[-1][p]+str(vars[j]))
        del_orders_fi.append(tmp_del_order_fi)
        del_orders_wi.append(tmp_del_order_wi)
        del_orders_F.append(tmp_del_order_F)
        del_orders_Fi.append(tmp_del_order_Fi)
        del_orders_Wi.append(tmp_del_order_Wi)
        del_orders_Ki.append(tmp_del_order_Ki)
        notation_f.append(tmp_notation_f)
        notation_w.append(tmp_notation_w)
        notation_k.append(tmp_notation_k)
    if order > 0:
        del_orders_fi.insert(0,[str(fi)])
        del_orders_wi.insert(0,[str(wi)])
        del_orders_F.insert(0,[str(F)])
        notation_f.insert(0,['f'])
        notation_w.insert(0,['w'])
        notation_k.insert(0,['k'])
    # Translate Derivatives into Code
    key = {str(rr) :'rr',
           str(rc):'rc',
           'tanh'  :'np.tanh',
           fi_inner:'f'}
    var = {'A' :'sol_mat[active,:,0]',
           'B1':'sol_mat[active,:,1]',
           'B2':'sol_mat[active,:,2]',
           'B3':'sol_mat[active,:,3]',
           'C1':'b_coef[active,:,1]',
           'C2':'b_coef[active,:,2]',
           'C3':'b_coef[active,:,3]',
           'C4':'b_coef[active,:,0]',
           'xc':'patch_1[active]',
           'yc':'patch_2[active]',
           'zc':'patch_3[active]',
           'xi':'patch_points[active,:,0]',
           'yi':'patch_points[active,:,1]',
           'zi':'patch_points[active,:,2]',
           'x' :'pt_3[:,:,0]',
           'y' :'pt_3[:,:,1]',
           'z' :'pt_3[:,:,2]',}
    funcs = {str(Fi):'f',
             str(Wi):'w',
             str(Ki):'k'}
    for i in range(len(del_orders_Fi)):
        for j in range(len(del_orders_Fi[i])):
            if order > 0:
                key[str(del_orders_Fi[i][j])] = notation_f[i+1][j]
                key[str(del_orders_Wi[i][j])] = notation_w[i+1][j]
                key[str(del_orders_Ki[i][j])] = notation_k[i+1][j]
            else:
                key[str(del_orders_Fi[i][j])] = notation_f[i][j]
                key[str(del_orders_Wi[i][j])] = notation_w[i][j]
                key[str(del_orders_Ki[i][j])] = notation_k[i][j]
    for i in range(len(del_orders_F)):
        for j in range(len(del_orders_F[i])):
            del_orders_fi[i][j] = str(del_orders_fi[i][j])
            del_orders_wi[i][j] = str(del_orders_wi[i][j])
            del_orders_F[i][j]  = str(del_orders_F[i][j])
            for item in key.keys():
                if item in del_orders_fi[i][j]:
                    #if key[item] == 'f' and i == 0:
                    #    pass
                    #else:
                    del_orders_fi[i][j] = del_orders_fi[i][j].replace(item,key[item])
                if item in del_orders_wi[i][j]:
                    del_orders_wi[i][j] = del_orders_wi[i][j].replace(item,key[item])
                if item in del_orders_F[i][j]:
                    del_orders_F[i][j] = del_orders_F[i][j].replace(item,key[item])
            for item in var.keys():
                if item in del_orders_fi[i][j]:
                    del_orders_fi[i][j] = del_orders_fi[i][j].replace(item,var[item])
                if item in del_orders_wi[i][j]:
                    del_orders_wi[i][j] = del_orders_wi[i][j].replace(item,var[item])
                if item in fs:
                    fs = fs.replace(item,var[item])
    for i in range(len(del_orders_F)):
        for j in range(len(del_orders_F[i])):
            for item in funcs.keys():
                if item in del_orders_F[i][j]:
                    del_orders_F[i][j] = del_orders_F[i][j].replace(item,funcs[item])
    for i in range(len(del_orders_fi)):
        if i == 0:
            continue
        for j in range(len(del_orders_fi[i])):
            pairs = outter_parentheses(del_orders_fi[i][j])
            old_fi = del_orders_fi[i][j]
            fixed_func = ''+old_fi[:pairs[0][0]]
            for ii,p in enumerate(pairs):
                if old_fi[p[1]-4:p[1]-2] == "**" and 'tanh' not in old_fi[p[0]:p[1]]:
                    p_square_start = old_fi.find("(",p[0]+1,p[1]-1)
                    p_square_end   = old_fi.rfind(")",p[0]+1,p[1]-1)
                    fixed_func += old_fi[p[0]:p_square_start]+'np.sum(np.nan_to_num'
                    fixed_func += old_fi[p_square_start:p_square_end+1]+',axis=1)'
                    fixed_func += old_fi[p_square_end+1:p[1]]
                    #fixed_func += 'np.sum(np.nan_to_num'+old_fi[p[0]:p[1]-4]+'),axis=1)'+old_fi[p[1]-4:p[1]]
                elif 'tanh' not in old_fi[p[0]:p[1]] and (p[1]-p[0]) > 3:
                    fixed_func += 'np.sum(np.nan_to_num'+old_fi[p[0]:p[1]]+',axis=1)'
                else:
                    fixed_func += old_fi[p[0]:p[1]]
                if ii < len(pairs)-1:
                    fixed_func += old_fi[p[1]:pairs[ii+1][0]]
                else:
                    if p[1] < len(old_fi):
                        fixed_func += old_fi[p[1]:]
            del_orders_fi[i][j] = fixed_func
    return del_orders_fi,del_orders_wi,del_orders_F,notation_f,notation_w,notation_k,fs

def construct(order,dim=3):
    funcs,weights,F,f_note,w_note,k_note,fs = d(order)
    functions = []
    func_strings = []
    for n in range(len(funcs)):
        d_func = 'def d_{}(xyzk,KDTree=None,'.format(n)+\
                           'patch_points=None,b_coef=None,'+\
                           'h=None,q=None,sol_mat=None,patch_1=None,'+\
                           'patch_2=None,patch_3=None,noskip=True):\n'
        d_func += '    x,y,z,k = xyzk\n'
        d_func += '    if noskip:\n'
        d_func += '        _,active = KDTree.query([x,y,z],k=k)\n'
        d_func += '    else:\n'
        d_func += '        active = list(range(len(patch_1)))\n'
        d_func += '    pt = np.ones((len(patch_1[active]),b_coef.shape[-1]))\n'
        d_func += '    pt[:,1] *= x\n'
        d_func += '    pt[:,2] *= y\n'
        d_func += '    pt[:,3] *= z\n'
        d_func += '    rc = ((x-patch_1[active])**2+(y-patch_2[active])**2+(z-patch_3[active])**2)**(1/2)\n'
        d_func += '    pt_3 = np.array([[[x,y,z]]])\n'
        d_func += '    if len(pt_3.shape) > 3:\n'
        d_func += '        pt_3 = pt_3.reshape(1,1,3)\n'
        #d_func += '    pt_3.flatten()
        d_func += '    rr = np.sum(np.square(pt_3-patch_points[active,:]),axis=2)\n'
        if n == 0:
            d_func += '    {} = {}\n'.format(f_note[0][0],fs)
            d_func += '    {} = np.nan_to_num({})\n'.format(f_note[0][0],f_note[0][0])
            d_func += '    {} = np.sum({},axis=1)\n'.format(f_note[0][0],f_note[0][0])
        elif n > 0:
            funcs[0][0] = fs
        for i in range(n+1):
            for j in range(len(weights[i])):
                d_func += '    {} = {}\n'.format(w_note[i][j],weights[i][j])
                d_func += '    {} = np.sum({})\n'.format(k_note[i][j],w_note[i][j])
        for i in range(n+1):
            for j in range(len(funcs[i])):
                tmp_func = funcs[i][j]
                d_func += '    {} = {}\n'.format(f_note[i][j],funcs[i][j])
                if i == 0 and n > 0:
                    d_func += '    {} = np.nan_to_num({})\n'.format(f_note[i][j],f_note[i][j])
                    d_func += '    {} = np.sum({},axis=1)\n'.format(f_note[i][j],f_note[i][j])
        d_func += '    res = np.zeros({})\n'.format(len(funcs[n]))
        if n > 0:
            d_func += '    {} = -np.tanh({})\n'.format(f_note[0][0],f_note[0][0])
        for j in range(len(funcs[n])):
            d_func += '    res_{} = {}\n'.format(f_note[n][j],F[n][j])
            d_func += '    res_{} = np.sum(res_{})\n'.format(f_note[n][j],f_note[n][j])
            d_func += '    res[{}] = res_{}\n'.format(j,f_note[n][j])
        if n > 1:
            d_func += '    res = res.reshape({}'.format(dim)
            for c in range(n-1):
                d_func += ',{}'.format(dim)
            d_func += ')\n'
        d_func += '    return res\n'
        #d_func += 'functions.append(d_{})'.format(n)
        exec(d_func)
        func_strings.append(d_func)
    return functions,func_strings

def gradient(xyzk,KDTree=None,patch_points=None,b_coef=None,h=None,q=None,sol_mat=None,patch_x=None,patch_y=None,patch_z=None):
    x,y,z,k = xyzk
    _,active = KDTree.query([x,y,z],k=k)
    pt = np.ones((len(patch_x[active]),len(b_coef[0])))
    pt[:,1] *= x
    pt[:,2] *= y
    pt[:,3] *= z
    r = ((x-patch_x[active])**2+(y-patch_y[active])**2+(z-patch_z[active])**2)**(1/2)
    weights = (1-np.tanh(r))/(r+h)**q
    dwx  = (-q*(1 - np.tanh(r))*(x-patch_x[active])*(h+r)**(-q-1))/r - ((1/np.cosh(r)**2)*(x-patch_x[active])*((h + r)**(-q)))/r
    dwy  = (-q*(1 - np.tanh(r))*(y-patch_y[active])*(h+r)**(-q-1))/r - ((1/np.cosh(r)**2)*(y-patch_y[active])*((h + r)**(-q)))/r
    dwz  = (-q*(1 - np.tanh(r))*(z-patch_z[active])*(h+r)**(-q-1))/r - ((1/np.cosh(r)**2)*(z-patch_z[active])*((h + r)**(-q)))/r
    pt_3 = np.array([[[x,y,z]]])
    scale = np.sqrt(np.sum(np.square(pt_3 - patch_points[active,:]),axis=2))
    fa = scale**3
    xcom = 3*scale*(pt_3[:,:,0] - patch_points[active,:,0])
    ycom = 3*scale*(pt_3[:,:,1] - patch_points[active,:,1])
    zcom = 3*scale*(pt_3[:,:,2] - patch_points[active,:,2])
    dx_1 = 3*scale*(pt_3[:,:,0] - patch_points[active,:,0])
    dx_2 = 3*(((pt_3[:,:,0] - patch_points[active,:,0])**2)/scale + scale)
    dx_3 = 3*(((pt_3[:,:,0] - patch_points[active,:,0])*(pt_3[:,:,1] - patch_points[active,:,1]))/scale)
    dx_4 = 3*(((pt_3[:,:,0] - patch_points[active,:,0])*(pt_3[:,:,2] - patch_points[active,:,2]))/scale)
    dy_1 = 3*scale*(pt_3[:,:,1] - patch_points[active,:,1])
    dy_2 = 3*(((pt_3[:,:,0] - patch_points[active,:,0])*(pt_3[:,:,1] - patch_points[active,:,1]))/scale)
    dy_3 = 3*(((pt_3[:,:,1] - patch_points[active,:,1])**2)/scale+scale)
    dy_4 = 3*(((pt_3[:,:,1] - patch_points[active,:,1])*(pt_3[:,:,2] - patch_points[active,:,2]))/scale)
    dz_1 = 3*scale*(pt_3[:,:,2] - patch_points[active,:,2])
    dz_2 = 3*(((pt_3[:,:,0] - patch_points[active,:,0])*(pt_3[:,:,2] - patch_points[active,:,2]))/scale)
    dz_3 = 3*(((pt_3[:,:,1] - patch_points[active,:,1])*(pt_3[:,:,2] - patch_points[active,:,2]))/scale)
    dz_4 = 3*(((pt_3[:,:,2] - patch_points[active,:,2])**2)/scale+scale)
    b = np.sum(pt*b_coef[active],axis=1)
    mat = np.zeros(sol_mat[active,:,:].shape)
    dx_mat = np.zeros(sol_mat[active,:,:].shape)
    dy_mat = np.zeros(sol_mat[active,:,:].shape)
    dz_mat = np.zeros(sol_mat[active,:,:].shape)
    mat[:,:,0] = sol_mat[active,:,0]*fa
    mat[:,:,1] = sol_mat[active,:,1]*xcom
    mat[:,:,2] = sol_mat[active,:,2]*ycom
    mat[:,:,3] = sol_mat[active,:,3]*zcom
    dx_mat[:,:,0] = sol_mat[active,:,0]*dx_1
    dx_mat[:,:,1] = sol_mat[active,:,1]*dx_2
    dx_mat[:,:,2] = sol_mat[active,:,2]*dx_3
    dx_mat[:,:,3] = sol_mat[active,:,3]*dx_4
    dy_mat[:,:,0] = sol_mat[active,:,0]*dy_1
    dy_mat[:,:,1] = sol_mat[active,:,1]*dy_2
    dy_mat[:,:,2] = sol_mat[active,:,2]*dy_3
    dy_mat[:,:,3] = sol_mat[active,:,3]*dy_4
    dz_mat[:,:,0] = sol_mat[active,:,0]*dz_1
    dz_mat[:,:,1] = sol_mat[active,:,1]*dz_2
    dz_mat[:,:,2] = sol_mat[active,:,2]*dz_3
    dz_mat[:,:,3] = sol_mat[active,:,3]*dz_4
    mat = np.nan_to_num(mat)
    dx_mat = np.nan_to_num(dx_mat)
    dy_mat = np.nan_to_num(dy_mat)
    dz_mat = np.nan_to_num(dz_mat)
    fsg = np.sum(np.sum(mat,axis=2),axis=1)+b
    total = np.sum(weights)
    dwx_total = np.sum(dwx)
    dwy_total = np.sum(dwy)
    dwz_total = np.sum(dwz)
    dx = -(1/np.cosh(fsg)**2)*(np.sum(np.sum(dx_mat,axis=2),axis=1)+b_coef[active,1])
    dy = -(1/np.cosh(fsg)**2)*(np.sum(np.sum(dy_mat,axis=2),axis=1)+b_coef[active,2])
    dz = -(1/np.cosh(fsg)**2)*(np.sum(np.sum(dz_mat,axis=2),axis=1)+b_coef[active,3])
    dx2 = ((-dwx*np.tanh(fsg)+dx*weights)*total+(dwx_total*weights*np.tanh(fsg)))/total**2
    dy2 = ((-dwy*np.tanh(fsg)+dy*weights)*total+(dwy_total*weights*np.tanh(fsg)))/total**2
    dz2 = ((-dwz*np.tanh(fsg)+dz*weights)*total+(dwz_total*weights*np.tanh(fsg)))/total**2
    #dx2 = -(1/np.cosh(fsg)**2)*(np.sum(np.sum(dx_mat,axis=2),axis=1)+b_coef[active,1])*(weights/total)#-(dwx/dwx_total)*np.tanh(fsg)
    #dy2 = -(1/np.cosh(fsg)**2)*(np.sum(np.sum(dy_mat,axis=2),axis=1)+b_coef[active,2])*(weights/total)#-(dwy/dwy_total)*np.tanh(fsg)
    #dz2 = -(1/np.cosh(fsg)**2)*(np.sum(np.sum(dz_mat,axis=2),axis=1)+b_coef[active,3])*(weights/total)#-(dwz/dwz_total)*np.tanh(fsg)
    #res_x = np.sum(weights*dx)/total
    #res_y = np.sum(weights*dy)/total
    #res_z = np.sum(weights*dz)/total
    res2_x = np.sum(dx2)
    res2_y = np.sum(dy2)
    res2_z = np.sum(dz2)
    return res2_x,res2_y,res2_z
def function_marching(x,y,z,patch_points=None,a_coef=None,b_coef=None,patch_x=None,patch_y=None,patch_z=None,q=None,h=None):
    interps = []
    weights = []
    total = 0
    for i in range(len(patch_x)):
        a = a_coef[i]
        b = b_coef[i]
        n = len(patch_points[i])
        xi = patch_points[i]
        fsg = b[0]+x*b[1]+y*b[2]+z*b[3]
        for j in range(n):
            scale = ((x-xi[j,0])**2+(y-xi[j,1])**2+(z-xi[j,2])**2)**(1/2)
            fsg += a[j]*(((x-xi[j,0])**2+(y-xi[j,1])**2+(z-xi[j,2])**2)**(1/2))**3
            fsg += a[n+j]*3*scale*(x-xi[j,0])
            fsg += a[n+j+n]*3*scale*(y-xi[j,1])
            fsg += a[n+j+2*n]*3*scale*(z-xi[j,2])
        fsg = -np.tanh(fsg)
        interps.append(fsg)
        r = ((x-patch_x[i])**2+(y-patch_y[i])**2+(z-patch_z[i])**2)**(1/2)
        value = (1-np.tanh(r))/(r+h)**q
        weights.append(value)
    total = sum(weights)
    res = 0
    for i in range(len(patch_x)):
        res += (weights[i]/total)*interps[i]
    return res

def function(xyzk,KDTree=None,patch_points=None,b_coef=None, h=None,q=None,sol_mat=None,patch_x=None,patch_y=None,patch_z=None):
    x,y,z,k = xyzk
    _,active = KDTree.query([x,y,z],k=k)
    pt = np.ones((len(patch_x[active]),len(b_coef[0])))
    pt[:,1] *= x
    pt[:,2] *= y
    pt[:,3] *= z
    r = ((x-patch_x[active])**2+(y-patch_y[active])**2+(z-patch_z[active])**2)**(1/2)
    weights = (1-np.tanh(r))/(r+h)**q
    pt_3 = np.array([[[x,y,z]]])
    scale = np.sqrt(np.sum(np.square(pt_3 - patch_points[active,:]),axis=2))
    fa = scale**3
    dx = 3*scale*(pt_3[:,:,0] - patch_points[active,:,0])
    dy = 3*scale*(pt_3[:,:,1] - patch_points[active,:,1])
    dz = 3*scale*(pt_3[:,:,2] - patch_points[active,:,2])
    b = np.sum(pt*b_coef[active],axis=1)
    mat = np.zeros(sol_mat[active,:,:].shape)
    mat[:,:,0] = sol_mat[active,:,0]*fa
    mat[:,:,1] = sol_mat[active,:,1]*dx
    mat[:,:,2] = sol_mat[active,:,2]*dy
    mat[:,:,3] = sol_mat[active,:,3]*dz
    mat = np.nan_to_num(mat)
    interps = -np.tanh(np.sum(np.sum(mat,axis=2),axis=1)+b)
    total = np.sum(weights)
    weights = weights/total
    res = np.sum(weights*interps)
    return res

def hessian(xyzk,KDTree=None,patch_points=None,b_coef=None,h=None,q=None,sol_mat=None,patch_x=None,patch_y=None,patch_z=None):
    x,y,z,k = xyzk
    _,active = KDTree.query([x,y,z],k=k)
    pt = np.ones((len(patch_x[active]),len(b_coef[0])))
    pt[:,1] *= x
    pt[:,2] *= y
    pt[:,3] *= z
    r = ((x-patch_x[active])**2+(y-patch_y[active])**2+(z-patch_z[active])**2)**(1/2)
    weights = (1-np.tanh(r))/(r+h)**q
    dwx  = (-q*(1 - np.tanh(r))*(x-patch_x[active])*(h+r)**(-q-1))/r - \
           ((1/np.cosh(r)**2)*(x-patch_x[active])*((h + r)**(-q)))/r
    dwy  = (-q*(1 - np.tanh(r))*(y-patch_y[active])*(h+r)**(-q-1))/r - \
           ((1/np.cosh(r)**2)*(y-patch_y[active])*((h + r)**(-q)))/r
    dwz  = (-q*(1 - np.tanh(r))*(z-patch_z[active])*(h+r)**(-q-1))/r - \
           ((1/np.cosh(r)**2)*(z-patch_z[active])*((h + r)**(-q)))/r
    dwxx = q**2*(1-np.tanh(r))*(-patch_x[active]+x)**2/((h+r)**2*(h+ r)**q*r**2) +\
           q*(1 -np.tanh(r))*(-patch_x[active] + x)**2/((h + r)**2*(h + r)**q*r**2) -\
           q*(1 -np.tanh(r))*(-patch_x[active] + x)*(patch_x[active]-x)/((h + r)*(h + r)**q*r**3) -\
           1.0*q*(1 -np.tanh(r))/((h + r)*(h + r)**q*r) + \
           2*q*(1 -np.tanh(r)**2)*(-patch_x[active] + x)**2/((h + r)*(h + r)**q*r**2) +\
           2*(1 -np.tanh(r)**2)*(-patch_x[active] + x)**2*np.tanh(r)/((h + r)**q*r**2) -\
           (1 -np.tanh(r)**2)*(-patch_x[active] + x)*(patch_x[active]-x)/((h + r)**q*r**3) -\
           1.0*(1 -np.tanh(r)**2)/((h + r)**q*r)
    #dwxx_2 = -((-q - 1)*q*(x - patch_x[active])**2*(1 - np.tanh(r))*(r + h)**(-q - 2))/r**2 -\
    #          (q*(1 - np.tanh(r))*(r + h)**(-q - 1))/r +\
    #          (q*(x - patch_x[active])**2*(1 - np.tanh(r))*(r + h)**(-q - 1))/r**3 +\
    #          (2*q*(x - patch_x[active])**2*(1/np.cosh(r)**2)*(r + h)**(-q - 1))/r**2 -\
    #          ((1/np.cosh(r)**2)*(r + h)**(-q))/r + ((x - patch_x[active])**2*(1/np.cosh(r)**2)*(r + h)**(-q))/r**3 +\
    #          (2*(x - patch_x[active])**2*np.tanh(r)*(1/np.cosh(r)**2)*(r + h)**(-q))/r**2
    #print('dwxx: {}'.format(dwxx[1]))
    #print('dwxx_2: {}'.format(dwxx_2[1]))
    dwxy = q**2*(1 -np.tanh(r))*(-patch_x[active]+x)*(-patch_y[active]+y)/((h + r)**2*(h + r)**q*r**2) +\
           q*(1 -np.tanh(r))*(-patch_x[active]+x)*(-patch_y[active]+y)/((h + r)**2*(h + r)**q*r**2) -\
           q*(1 -np.tanh(r))*(-patch_x[active]+x)*(patch_y[active]-y)/((h + r)*(h + r)**q*r**3) +\
           2*q*(1 -np.tanh(r)**2)*(-patch_x[active]+x)*(-patch_y[active]+y)/((h + r)*(h + r)**q*r**2) +\
           2*(1 -np.tanh(r)**2)*(-patch_x[active]+x)*(-patch_y[active]+y)*np.tanh(r)/((h + r)**q*r**2) -\
           (1 -np.tanh(r)**2)*(-patch_x[active]+x)*(patch_y[active] - y)/((h + r)**q*r**3)
    dwxz = q**2*(1 -np.tanh(r))*(-patch_x[active]+x)*(-patch_z[active]+z)/((h + r)**2*(h + r)**q*r**2) +\
           q*(1 -np.tanh(r))*(-patch_x[active]+x)*(-patch_z[active]+z)/((h + r)**2*(h + r)**q*r**2) -\
           q*(1 -np.tanh(r))*(-patch_x[active]+x)*(patch_z[active]-z)/((h + r)*(h + r)**q*r**3) +\
           2*q*(1 -np.tanh(r)**2)*(-patch_x[active]+x)*(-patch_z[active]+z)/((h + r)*(h + r)**q*r**2) +\
           2*(1 -np.tanh(r)**2)*(-patch_x[active]+x)*(-patch_z[active]+z)*np.tanh(r)/((h + r)**q*r**2) -\
           (1 -np.tanh(r)**2)*(-patch_x[active]+x)*(patch_z[active]-z)/((h + r)**q*r**3)
    dwyx = dwxy
    dwyy = q**2*(1 -np.tanh(r))*(-patch_y[active] + y)**2/((h + r)**2*(h + r)**q*r**2) +\
           q*(1 -np.tanh(r))*(-patch_y[active] + y)**2/((h + r)**2*(h + r)**q*r**2) -\
           q*(1 -np.tanh(r))*(-patch_y[active] + y)*(patch_y[active] - y)/((h + r)*(h + r)**q*r**3) -\
           1.0*q*(1 -np.tanh(r))/((h + r)*(h + r)**q*r) +\
           2*q*(1 -np.tanh(r)**2)*(-patch_y[active] + y)**2/((h + r)*(h + r)**q*r**2) +\
           2*(1 -np.tanh(r)**2)*(-patch_y[active] + y)**2*np.tanh(r)/((h + r)**q*r**2) -\
           (1 -np.tanh(r)**2)*(-patch_y[active] + y)*(patch_y[active] - y)/((h + r)**q*r**3) -\
           1.0*(1 -np.tanh(r)**2)/((h + r)**q*r)
    dwyz = q**2*(1 -np.tanh(r))*(-patch_y[active] + y)*(-patch_z[active] + z)/((h + r)**2*(h + r)**q*r**2) +\
           q*(1 -np.tanh(r))*(-patch_y[active] + y)*(-patch_z[active] + z)/((h + r)**2*(h + r)**q*r**2) -\
           q*(1 -np.tanh(r))*(-patch_y[active] + y)*(patch_z[active] - z)/((h + r)*(h + r)**q*r**3) +\
           2*q*(1 -np.tanh(r)**2)*(-patch_y[active] + y)*(-patch_z[active] + z)/((h + r)*(h + r)**q*r**2) +\
           2*(1 -np.tanh(r)**2)*(-patch_y[active] + y)*(-patch_z[active] + z)*np.tanh(r)/((h + r)**q*r**2) -\
           (1 -np.tanh(r)**2)*(-patch_y[active] + y)*(patch_z[active] - z)/((h + r)**q*r**3)
    dwzx = dwxz
    dwzy = dwyz
    dwzz = q**2*(1 -np.tanh(r))*(-patch_z[active] + z)**2/((h + r)**2*(h + r)**q*r**2) +\
           q*(1 -np.tanh(r))*(-patch_z[active] + z)**2/((h + r)**2*(h + r)**q*r**2) -\
           q*(1 -np.tanh(r))*(-patch_z[active] + z)*(patch_z[active] - z)/((h + r)*(h + r)**q*r**3) -\
           1.0*q*(1 -np.tanh(r))/((h + r)*(h + r)**q*r) +\
           2*q*(1 -np.tanh(r)**2)*(-patch_z[active] + z)**2/((h + r)*(h + r)**q*r**2) +\
           2*(1 -np.tanh(r)**2)*(-patch_z[active] + z)**2*np.tanh(r)/((h + r)**q*r**2) -\
           (1 -np.tanh(r)**2)*(-patch_z[active] + z)*(patch_z[active] - z)/((h + r)**q*r**3) -\
           1.0*(1 -np.tanh(r)**2)/((h + r)**q*r)
    pt_3 = np.array([[[x,y,z]]])
    scale = np.sqrt(np.sum(np.square(pt_3 - patch_points[active,:]),axis=2))
    fa = scale**3
    xcom = 3*scale*(pt_3[:,:,0] - patch_points[active,:,0])
    ycom = 3*scale*(pt_3[:,:,1] - patch_points[active,:,1])
    zcom = 3*scale*(pt_3[:,:,2] - patch_points[active,:,2])
    dx_1 = 3*scale*(pt_3[:,:,0] - patch_points[active,:,0])
    dx_2 = 3*(((pt_3[:,:,0] - patch_points[active,:,0])**2)/scale + scale)
    dx_3 = 3*(((pt_3[:,:,0] - patch_points[active,:,0])*(pt_3[:,:,1] - patch_points[active,:,1]))/scale)
    dx_4 = 3*(((pt_3[:,:,0] - patch_points[active,:,0])*(pt_3[:,:,2] - patch_points[active,:,2]))/scale)
    dy_1 = 3*scale*(pt_3[:,:,1] - patch_points[active,:,1])
    dy_2 = 3*(((pt_3[:,:,0] - patch_points[active,:,0])*(pt_3[:,:,1] - patch_points[active,:,1]))/scale)
    dy_3 = 3*(((pt_3[:,:,1] - patch_points[active,:,1])**2)/scale+scale)
    dy_4 = 3*(((pt_3[:,:,1] - patch_points[active,:,1])*(pt_3[:,:,2] - patch_points[active,:,2]))/scale)
    dz_1 = 3*scale*(pt_3[:,:,2] - patch_points[active,:,2])
    dz_2 = 3*(((pt_3[:,:,0] - patch_points[active,:,0])*(pt_3[:,:,2] - patch_points[active,:,2]))/scale)
    dz_3 = 3*(((pt_3[:,:,1] - patch_points[active,:,1])*(pt_3[:,:,2] - patch_points[active,:,2]))/scale)
    dz_4 = 3*(((pt_3[:,:,2] - patch_points[active,:,2])**2)/scale+scale)
    b = np.sum(pt*b_coef[active],axis=1)
    mat = np.zeros(sol_mat[active,:,:].shape)
    dx_mat = np.zeros(sol_mat[active,:,:].shape)
    dy_mat = np.zeros(sol_mat[active,:,:].shape)
    dz_mat = np.zeros(sol_mat[active,:,:].shape)
    mat[:,:,0] = sol_mat[active,:,0]*fa
    mat[:,:,1] = sol_mat[active,:,1]*xcom
    mat[:,:,2] = sol_mat[active,:,2]*ycom
    mat[:,:,3] = sol_mat[active,:,3]*zcom
    dx_mat[:,:,0] = sol_mat[active,:,0]*dx_1
    dx_mat[:,:,1] = sol_mat[active,:,1]*dx_2
    dx_mat[:,:,2] = sol_mat[active,:,2]*dx_3
    dx_mat[:,:,3] = sol_mat[active,:,3]*dx_4
    dy_mat[:,:,0] = sol_mat[active,:,0]*dy_1
    dy_mat[:,:,1] = sol_mat[active,:,1]*dy_2
    dy_mat[:,:,2] = sol_mat[active,:,2]*dy_3
    dy_mat[:,:,3] = sol_mat[active,:,3]*dy_4
    dz_mat[:,:,0] = sol_mat[active,:,0]*dz_1
    dz_mat[:,:,1] = sol_mat[active,:,1]*dz_2
    dz_mat[:,:,2] = sol_mat[active,:,2]*dz_3
    dz_mat[:,:,3] = sol_mat[active,:,3]*dz_4
    mat = np.nan_to_num(mat)
    dx_mat = np.nan_to_num(dx_mat)
    dy_mat = np.nan_to_num(dy_mat)
    dz_mat = np.nan_to_num(dz_mat)
    #dxx
    dxx_1 = (pt_3[:,:,0] - patch_points[active,:,0])*(3*(pt_3[:,:,0] - patch_points[active,:,0])*(patch_points[active,:,0]-pt_3[:,:,0]))/scale**3 #B1
    dxx_2 = (pt_3[:,:,1] - patch_points[active,:,1])*(3*(pt_3[:,:,0] - patch_points[active,:,0])*(patch_points[active,:,0]-pt_3[:,:,0]))/scale**3 #B2
    dxx_3 = (pt_3[:,:,2] - patch_points[active,:,2])*(3*(pt_3[:,:,0] - patch_points[active,:,0])*(patch_points[active,:,0]-pt_3[:,:,0]))/scale**3 #B3
    dxx_4 = 6*(pt_3[:,:,0] - patch_points[active,:,0])/scale #B1
    dxx_5 = (3*(pt_3[:,:,0] - patch_points[active,:,0])**2)/scale #A
    dxx_6 = 3*(pt_3[:,:,0] - patch_points[active,:,0])/scale #B1
    dxx_7 = 3*(pt_3[:,:,1] - patch_points[active,:,1])/scale #B2
    dxx_8 = 3*(pt_3[:,:,2] - patch_points[active,:,2])/scale #B3
    dxx_9 = 3*scale #A
    #dxy
    dxy_1 = (pt_3[:,:,0]-patch_points[active,:,0])*(3*(pt_3[:,:,0]-patch_points[active,:,0])*(patch_points[active,:,1] - pt_3[:,:,1]))/scale**3 #B1
    dxy_2 = (pt_3[:,:,1]-patch_points[active,:,1])*(3*(pt_3[:,:,0]-patch_points[active,:,0])*(patch_points[active,:,1] - pt_3[:,:,1]))/scale**3 #B2
    dxy_3 = (pt_3[:,:,2]-patch_points[active,:,2])*(3*(pt_3[:,:,0]-patch_points[active,:,0])*(patch_points[active,:,1] - pt_3[:,:,1]))/scale**3 #B3
    dxy_4 = 3*(pt_3[:,:,0] - patch_points[active,:,0])/scale #B2
    dxy_5 = 3*(pt_3[:,:,1] - patch_points[active,:,1])/scale #B1
    dxy_6 = (3*(pt_3[:,:,0] - patch_points[active,:,0])*(pt_3[:,:,1] - patch_points[active,:,1]))/scale #A
    # dxz
    dxz_1 = (pt_3[:,:,0]-patch_points[active,:,0])*(3*(pt_3[:,:,0]-patch_points[active,:,0])*(patch_points[active,:,2]-pt_3[:,:,2]))/scale**3 #B1
    dxz_2 = (pt_3[:,:,1]-patch_points[active,:,1])*(3*(pt_3[:,:,0]-patch_points[active,:,0])*(patch_points[active,:,2]-pt_3[:,:,2]))/scale**3 #B2
    dxz_3 = (pt_3[:,:,2]-patch_points[active,:,2])*(3*(pt_3[:,:,0]-patch_points[active,:,0])*(patch_points[active,:,2]-pt_3[:,:,2]))/scale**3 #B3
    dxz_4 = 3*(pt_3[:,:,0] - patch_points[active,:,0])/scale #B3
    dxz_5 = 3*(pt_3[:,:,2] - patch_points[active,:,2])/scale #B1
    dxz_6 = (3*(pt_3[:,:,0] - patch_points[active,:,0])*(pt_3[:,:,2] - patch_points[active,:,2]))/scale #A
    #dyy
    dyy_1 = (pt_3[:,:,0] - patch_points[active,:,0])*(3*(pt_3[:,:,1] - patch_points[active,:,1])*(patch_points[active,:,1] - pt_3[:,:,1]))/scale**3 #B1
    dyy_2 = (pt_3[:,:,1] - patch_points[active,:,1])*(3*(pt_3[:,:,1] - patch_points[active,:,1])*(patch_points[active,:,1] - pt_3[:,:,1]))/scale**3 #B2
    dyy_3 = (pt_3[:,:,2] - patch_points[active,:,2])*(3*(pt_3[:,:,1] - patch_points[active,:,1])*(patch_points[active,:,1] - pt_3[:,:,1]))/scale**3 #B3
    dyy_4 = 6*(pt_3[:,:,1] - patch_points[active,:,1])/scale #B2
    dyy_5 = (3*(pt_3[:,:,1] - patch_points[active,:,1])**2)/scale #A
    dyy_6 = 3*(pt_3[:,:,0] - patch_points[active,:,0])/scale #B1
    dyy_7 = 3*(pt_3[:,:,1] - patch_points[active,:,1])/scale #B2
    dyy_8 = 3*(pt_3[:,:,2] - patch_points[active,:,2])/scale #B3
    dyy_9 = 3*scale #A
    #dyz
    dyz_1 = (pt_3[:,:,0]-patch_points[active,:,0])*(3*(pt_3[:,:,1]-patch_points[active,:,1])*(patch_points[active,:,2]-pt_3[:,:,2]))/scale**3 #B1
    dyz_2 = (pt_3[:,:,1]-patch_points[active,:,1])*(3*(pt_3[:,:,1]-patch_points[active,:,1])*(patch_points[active,:,2]-pt_3[:,:,2]))/scale**3 #B2
    dyz_3 = (pt_3[:,:,2]-patch_points[active,:,2])*(3*(pt_3[:,:,1]-patch_points[active,:,1])*(patch_points[active,:,2]-pt_3[:,:,2]))/scale**3 #B3
    dyz_4 = 3*(pt_3[:,:,1] - patch_points[active,:,1])/scale #B3
    dyz_5 = 3*(pt_3[:,:,2] - patch_points[active,:,2])/scale #B2
    dyz_6 = (3*(pt_3[:,:,1] - patch_points[active,:,1])*(pt_3[:,:,2] - patch_points[active,:,2]))/scale #A
    #dzz
    dzz_1 = (pt_3[:,:,0] - patch_points[active,:,0])*(3*(pt_3[:,:,2] - patch_points[active,:,2])*(patch_points[active,:,2] - pt_3[:,:,2]))/scale**3 #B1
    dzz_2 = (pt_3[:,:,1] - patch_points[active,:,1])*(3*(pt_3[:,:,2] - patch_points[active,:,2])*(patch_points[active,:,2] - pt_3[:,:,2]))/scale**3 #B2
    dzz_3 = (pt_3[:,:,2] - patch_points[active,:,2])*(3*(pt_3[:,:,2] - patch_points[active,:,2])*(patch_points[active,:,2] - pt_3[:,:,2]))/scale**3 #B3
    dzz_4 = 6*(pt_3[:,:,2] - patch_points[active,:,2])/scale #B3
    dzz_5 = (3*(pt_3[:,:,2] - patch_points[active,:,2])**2)/scale #A
    dzz_6 = 3*(pt_3[:,:,0] - patch_points[active,:,0])/scale #B1
    dzz_7 = 3*(pt_3[:,:,1] - patch_points[active,:,1])/scale #B2
    dzz_8 = 3*(pt_3[:,:,2] - patch_points[active,:,2])/scale #B3
    dzz_9 = 3*scale #A
    dxx_mat = np.zeros(sol_mat[active,:,:].shape)
    dxy_mat = np.zeros(sol_mat[active,:,:].shape)
    dxz_mat = np.zeros(sol_mat[active,:,:].shape)
    dyy_mat = np.zeros(sol_mat[active,:,:].shape)
    dyz_mat = np.zeros(sol_mat[active,:,:].shape)
    dzz_mat = np.zeros(sol_mat[active,:,:].shape)
    #multiply solution coefficients to variable values
    dxx_mat[:,:,0] = sol_mat[active,:,0]*(dxx_5+dxx_9)
    dxx_mat[:,:,1] = sol_mat[active,:,1]*(dxx_1+dxx_4+dxx_6)
    dxx_mat[:,:,2] = sol_mat[active,:,2]*(dxx_2+dxx_7)
    dxx_mat[:,:,3] = sol_mat[active,:,3]*(dxx_3+dxx_8)

    dxy_mat[:,:,0] = sol_mat[active,:,0]*(dxy_6)
    dxy_mat[:,:,1] = sol_mat[active,:,1]*(dxy_1+dxy_5)
    dxy_mat[:,:,2] = sol_mat[active,:,2]*(dxy_2+dxy_4)
    dxy_mat[:,:,3] = sol_mat[active,:,3]*(dxy_3)

    dxz_mat[:,:,0] = sol_mat[active,:,0]*(dxz_6)
    dxz_mat[:,:,1] = sol_mat[active,:,1]*(dxz_1+dxz_5)
    dxz_mat[:,:,2] = sol_mat[active,:,2]*(dxz_2)
    dxz_mat[:,:,3] = sol_mat[active,:,3]*(dxz_3+dxz_4)

    dyy_mat[:,:,0] = sol_mat[active,:,0]*(dyy_5+dyy_9)
    dyy_mat[:,:,1] = sol_mat[active,:,1]*(dyy_1+dyy_6)
    dyy_mat[:,:,2] = sol_mat[active,:,2]*(dyy_2+dyy_4+dyy_7)
    dyy_mat[:,:,3] = sol_mat[active,:,3]*(dyy_3+dyy_8)

    dyz_mat[:,:,0] = sol_mat[active,:,0]*(dyz_6)
    dyz_mat[:,:,1] = sol_mat[active,:,1]*(dyz_1)
    dyz_mat[:,:,2] = sol_mat[active,:,2]*(dyz_2+dyz_5)
    dyz_mat[:,:,3] = sol_mat[active,:,3]*(dyz_3+dyz_4)

    dzz_mat[:,:,0] = sol_mat[active,:,0]*(dzz_5+dzz_9)
    dzz_mat[:,:,1] = sol_mat[active,:,1]*(dzz_1+dzz_6)
    dzz_mat[:,:,2] = sol_mat[active,:,2]*(dzz_2+dzz_7)
    dzz_mat[:,:,3] = sol_mat[active,:,3]*(dzz_3+dzz_4+dzz_8)
    dxx_mat = np.nan_to_num(dxx_mat)
    dxy_mat = np.nan_to_num(dxy_mat)
    dxz_mat = np.nan_to_num(dxz_mat)
    dyy_mat = np.nan_to_num(dyy_mat)
    dyz_mat = np.nan_to_num(dyz_mat)
    dzz_mat = np.nan_to_num(dzz_mat)
    #combine inner components to arrive at gradient, function, and partial
    #hessian values
    dxx_mat = np.sum(np.sum(dxx_mat,axis=2),axis=1)
    dxy_mat = np.sum(np.sum(dxy_mat,axis=2),axis=1)
    dxz_mat = np.sum(np.sum(dxz_mat,axis=2),axis=1)
    dyy_mat = np.sum(np.sum(dyy_mat,axis=2),axis=1)
    dyz_mat = np.sum(np.sum(dyz_mat,axis=2),axis=1)
    dzz_mat = np.sum(np.sum(dzz_mat,axis=2),axis=1)

    fsg = np.sum(np.sum(mat,axis=2),axis=1)+b
    dx2 = (np.sum(np.sum(dx_mat,axis=2),axis=1)+b_coef[active,1])
    dy2 = (np.sum(np.sum(dy_mat,axis=2),axis=1)+b_coef[active,2])
    dz2 = (np.sum(np.sum(dz_mat,axis=2),axis=1)+b_coef[active,3])
    dx = -(1/np.cosh(fsg)**2)*(np.sum(np.sum(dx_mat,axis=2),axis=1)+b_coef[active,1])
    dy = -(1/np.cosh(fsg)**2)*(np.sum(np.sum(dy_mat,axis=2),axis=1)+b_coef[active,2])
    dz = -(1/np.cosh(fsg)**2)*(np.sum(np.sum(dz_mat,axis=2),axis=1)+b_coef[active,3])
    dwx_total = np.sum(dwx)
    dwy_total = np.sum(dwy)
    dwz_total = np.sum(dwz)
    dwxx_total = np.sum(dwxx)
    dwxy_total = np.sum(dwxy)
    dwxz_total = np.sum(dwxz)
    dwyx_total = np.sum(dwyx)
    dwyy_total = np.sum(dwyy)
    dwyz_total = np.sum(dwyz)
    dwzx_total = np.sum(dwzx)
    dwzy_total = np.sum(dwzy)
    dwzz_total = np.sum(dwzz)
    total = np.sum(weights)
    dx22 = ((-dwx*np.tanh(fsg)+dx*weights)*total+(dwx_total*weights*np.tanh(fsg)))/total**2
    dy22 = ((-dwy*np.tanh(fsg)+dy*weights)*total+(dwy_total*weights*np.tanh(fsg)))/total**2
    dz22 = ((-dwz*np.tanh(fsg)+dz*weights)*total+(dwz_total*weights*np.tanh(fsg)))/total**2
    #print('dx: {}'.format(np.sum(dx22)))
    #print('dy: {}'.format(np.sum(dy22)))
    #print('dz: {}'.format(np.sum(dz22)))
    #dx2 = ((-dwx*np.tanh(fsg)+dx*weights)*total+(dwx_total*weights*np.tanh(fsg)))/total**2
    #dy2 = ((-dwy*np.tanh(fsg)+dy*weights)*total+(dwy_total*weights*np.tanh(fsg)))/total**2
    #dz2 = ((-dwz*np.tanh(fsg)+dz*weights)*total+(dwz_total*weights*np.tanh(fsg)))/total**2

    dxx = -dxx_mat*(1/np.cosh(fsg))**2+2*dx2**2*(1/np.cosh(fsg))**2*np.tanh(fsg)
    dxy = -dxy_mat*(1/np.cosh(fsg))**2+2*dx2*dy2*(1/np.cosh(fsg))**2*np.tanh(fsg)
    dxz = -dxz_mat*(1/np.cosh(fsg))**2+2*dx2*dz2*(1/np.cosh(fsg))**2*np.tanh(fsg)
    dyy = -dyy_mat*(1/np.cosh(fsg))**2+2*dy2**2*(1/np.cosh(fsg))**2*np.tanh(fsg)
    dyz = -dyz_mat*(1/np.cosh(fsg))**2+2*dy2*dz2*(1/np.cosh(fsg))**2*np.tanh(fsg)
    dzz = -dzz_mat*(1/np.cosh(fsg))**2+2*dz2**2*(1/np.cosh(fsg))**2*np.tanh(fsg)

    #res_dxx = (((-dwxx*np.tanh(fsg)+dx*dwx+dxx*weights+dx*dwx)*total+\
    #           dwx_total*(-dwx*np.tanh(fsg)+dx*weights)-\
    #           (dwxx_total*(-weights*np.tanh(fsg))+\
    #            dwx_total*(-dwx*np.tanh(fsg)+dx*weights)))*total**2 -\
    #           2*dwx_total*((-dwx*np.tanh(fsg)+dx*weights)*total - \
    #                         dwx_total*(-weights*np.tanh(fsg))))/total**4
    res_dxx = (-total*(weights*(2*dx*dwx_total + -np.tanh(fsg)*dwxx_total) +
               2*-np.tanh(fsg)*dwx_total*dwx) + total**2*(weights*dxx + 2*dx*dwx +
               -np.tanh(fsg)*dwxx) + 2*-np.tanh(fsg)*weights*dwx_total**2)/total**3
    #res_dxy = (((-dwxy*np.tanh(fsg)+dwx*dy+dxy*weights+dwy*dx)*total+\
    #           dwy_total*(-dwx*np.tanh(fsg)+dx*weights)-\
    #           (dwxy_total*(-weights*np.tanh(fsg))+\
    #            dwx_total*(-dwy*np.tanh(fsg)+dy*weights)))*total**2 -\
    #            2*dwy_total*((-dwx*np.tanh(fsg)+dx*weights)*total - \
    #                          dwx_total*(-weights*np.tanh(fsg))))/total**4
    res_dxy = (-total*(weights*(dx*dwy_total + dy*dwx*total + -np.tanh(fsg)*dwxy_total) +
               -np.tanh(fsg)*(dwx_total*dwy + dwy_total*dwy)) + total**2*(dx*dwy + dy*dwy +
               dxy*weights + -np.tanh(fsg)*dwxy) +
               2*-np.tanh(fsg)*dwy_total*dwx_total*weights)/total**3

    #res_dxz = (((-dwxz*np.tanh(fsg)+dwx*dz+dxz*weights+dwz*dx)*total+\
    #           dwz_total*(-dwx*np.tanh(fsg)+dx*weights)-\
    #           (dwxz_total*(-weights*np.tanh(fsg))+\
    #            dwx_total*(-dwz*np.tanh(fsg)+dz*weights)))*total**2 -\
    #            2*dwz_total*((-dwx*np.tanh(fsg)+dx*weights)*total - \
    #                          dwx_total*(-weights*np.tanh(fsg))))/total**4
    res_dxz = (-total*(weights*(dx*dwz_total + dz*dwx*total + -np.tanh(fsg)*dwxz_total) +
               -np.tanh(fsg)*(dwx_total*dwz + dwz_total*dwz)) + total**2*(dx*dwz + dz*dwz +
               dxz*weights + -np.tanh(fsg)*dwxz) +
               2*-np.tanh(fsg)*dwz_total*dwx_total*weights)/total**3
    #res_dyy = (((-dwyy*np.tanh(fsg)+dy*dwy+dyy*weights+dy*dwy)*total+\
    #           dwy_total*(-dwy*np.tanh(fsg)+dy*weights)-\
    #           (dwyy_total*(-weights*np.tanh(fsg))+\
    #            dwy_total*(-dwy*np.tanh(fsg)+dy*weights)))*total**2 -\
    #           2*dwy_total*((-dwy*np.tanh(fsg)+dy*weights)*total - \
    #                         dwy_total*(-weights*np.tanh(fsg))))/total**4
    res_dyy = (-total*(weights*(2*dy*dwy_total-np.tanh(fsg)*dwyy_total)-
               2*np.tanh(fsg)*dwy_total*dwy)+total**2*(weights*dyy+2*dy*dwy-
               np.tanh(fsg)*dwyy)-2*np.tanh(fsg)*weights*dwy_total**2)/total**3
    #res_dyz = (((-dwyz*np.tanh(fsg)+dwy*dz+dyz*weights+dwz*dy)*total+\
    #           dwz_total*(-dwy*np.tanh(fsg)+dy*weights)-\
    #           (dwyz_total*(-weights*np.tanh(fsg))+\
    #            dwy_total*(-dwz*np.tanh(fsg)+dz*weights)))*total**2 -\
    #            2*dwz_total*((-dwy*np.tanh(fsg)+dy*weights)*total - \
    #                          dwy_total*(-weights*np.tanh(fsg))))/total**4
    res_dyz = (-total*(weights*(dy*dwz_total + dz*dwy*total + -np.tanh(fsg)*dwyz_total) +
               -np.tanh(fsg)*(dwy_total*dwz + dwz_total*dwz)) + total**2*(dy*dwz + dz*dwz +
               dxz*weights + -np.tanh(fsg)*dwyz) +
               2*-np.tanh(fsg)*dwz_total*dwy_total*weights)/total**3
    #res_dzz = (((-dwzz*np.tanh(fsg)+dz*dwz+dzz*weights+dz*dwz)*total+\
    #            dwz_total*(-dwz*np.tanh(fsg)+dz*weights)-\
    #            (dwzz_total*(-weights*np.tanh(fsg))+\
    #             dwz_total*(-dwz*np.tanh(fsg)+dz*weights)))*total**2 -\
    #             2*dwz_total*((-dwz*np.tanh(fsg)+dz*weights)*total - \
    #                           dwz_total*(-weights*np.tanh(fsg))))/total**4
    res_dzz = (-total*(weights*(2*dz*dwz_total-np.tanh(fsg)*dwzz_total)-
               2*np.tanh(fsg)*dwz_total*dwz)+total**2*(weights*dzz+2*dz*dwz-
               np.tanh(fsg)*dwzz)-2*np.tanh(fsg)*weights*dwz_total**2)/total**3
    #total = np.sum(weights)
    #res_dxx = np.sum(weights*dxx)/total
    #res_dxy = np.sum(weights*dxy)/total
    #res_dxz = np.sum(weights*dxz)/total
    #res_dyy = np.sum(weights*dyy)/total
    #res_dyz = np.sum(weights*dyz)/total
    #res_dzz = np.sum(weights*dzz)/total
    res_dxx = np.sum(res_dxx)
    res_dxy = np.sum(res_dxy)
    res_dxz = np.sum(res_dxz)
    res_dyy = np.sum(res_dyy)
    res_dyz = np.sum(res_dyz)
    res_dzz = np.sum(res_dzz)
    return np.array([[res_dxx,res_dxy,res_dxz],[res_dxy,res_dyy,res_dyz],[res_dxz,res_dyz,res_dzz]])

def core_hessian(x,y,z,patch_points=None,a_coef=None,b_coef=None,patch_x=None,patch_y=None,patch_z=None,q=None,h=None):
    fsg_dxx = []
    fsg_dxy = []
    fsg_dxz = []
    fsg_dyy = []
    fsg_dyz = []
    fsg_dzz = []
    weights = []
    for i in range(len(patch_x)):
        a = a_coef[i]
        b = b_coef[i]
        n = len(patch_points[i])
        xi = patch_points[i]
        #patch = patches[i]
        dx = b[1]
        dy = b[2]
        dz = b[3]
        fsg = b[0]+x*b[1]+y*b[2]+z*b[3]
        dxx = 0
        dxy = 0
        dxz = 0
        dyy = 0
        dyz = 0
        dzz = 0
        for j in range(n):
            scale = ((x-xi[j,0])**2+(y-xi[j,1])**2+(z-xi[j,2])**2)**(1/2)
            #f_value = np.sech(a[j]*scale**3+3*scale*(a[n+j]*(x-xi[j,0])+(y-xi[j,1])+(z-xi[j,2]))+x*b[1]+y*b[2]+z*b[3]+b[0])**2
            # function value
            fsg += a[j]*(((x-xi[j,0])**2+(y-xi[j,1])**2+(z-xi[j,2])**2)**(1/2))**3
            fsg += a[n+j]*3*scale*(x-xi[j,0])
            fsg += a[n+j+n]*3*scale*(y-xi[j,1])
            fsg += a[n+j+2*n]*3*scale*(z-xi[j,2])
            # derivative values
            dx += a[j]*3*scale*(x-xi[j,0])
            dy += a[j]*3*scale*(y-xi[j,1])
            dz += a[j]*3*scale*(z-xi[j,2])
            # hessian x-components
            dxx += (-((3*(x-xi[j,0])**2*(a[n+j]*(x-xi[j,0])+a[n+j+n]*(y-xi[j,1])+a[n+j+2*n]*(z-xi[j,2])))/
                      (scale**3))+(6*a[n+j]*(x-xi[j,0]))/scale+(3*a[j]*(x-xi[j,0])**2)/scale+
                    (3*(a[n+j]*(x-xi[j,0])+a[n+j+n]*(y-xi[j,1])+a[n+j+2*n]*(z-xi[j,2])))/scale+
                    3*a[j]*scale)
            dxy += (-((3*(x-xi[j,0])*(y-xi[j,1])*(a[n+j]*(x-xi[j,0])+a[n+j+n]*(y-xi[j,1])+a[n+j+2*n]*(z-xi[j,2])))/
                      (scale**3))+(3*a[n+j+n]*(x-xi[j,0]))/scale+(3*a[n+j]*(y-xi[j,1]))/scale+
                    (3*(a[j]*(x-xi[j,0])*(y-xi[j,1])))/scale)
            dxz += (-((3*(x-xi[j,0])*(z-xi[j,2])*(a[n+j]*(x-xi[j,0])+a[n+j+n]*(y-xi[j,1])+a[n+j+2*n]*(z-xi[j,2])))/
                      (scale**3))+(3*a[n+j+2*n]*(x-xi[j,0]))/scale+(3*a[n+j]*(z-xi[j,2]))/scale+
                    (3*(a[j]*(x-xi[j,0])*(z-xi[j,2])))/scale)
            dyy += (-((3*(y-xi[j,1])**2*(a[n+j]*(x-xi[j,0])+a[n+j+n]*(y-xi[j,1])+a[n+j+2*n]*(z-xi[j,2])))/
                      (scale**3))+(6*a[n+j+n]*(y-xi[j,1]))/scale+(3*a[j]*(y-xi[j,1])**2)/scale+
                    (3*(a[n+j]*(x-xi[j,0])+a[n+j+n]*(y-xi[j,1])+a[n+j+2*n]*(z-xi[j,2])))/scale+
                    3*a[j]*scale)
            dyz += (-((3*(y-xi[j,1])*(z-xi[j,2])*(a[n+j]*(x-xi[j,0])+a[n+j+n]*(y-xi[j,1])+a[n+j+2*n]*(z-xi[j,2])))/
                      (scale**3))+(3*a[n+j+2*n]*(y-xi[j,1]))/scale+(3*a[n+j+n]*(z-xi[j,2]))/scale+
                    (3*(a[j]*(y-xi[j,1])*(z-xi[j,2])))/scale)
            dzz += (-((3*(z-xi[j,2])**2*(a[n+j]*(x-xi[j,0])+a[n+j+n]*(y-xi[j,1])+a[n+j+2*n]*(z-xi[j,2])))/
                      (scale**3))+(6*a[n+j+2*n]*(z-xi[j,2]))/scale+(3*a[j]*(z-xi[j,2])**2)/scale+
                    (3*(a[n+j]*(x-xi[j,0])+a[n+j+n]*(y-xi[j,1])+a[n+j+2*n]*(z-xi[j,2])))/scale+
                    3*a[j]*scale)
            #x-derivatives components numerator
            dx += a[n+j]*3*((((x-xi[j,0])**2)/scale)+scale)
            dx += a[n+j+n]*3*(((x-xi[j,0])*(y-xi[j,1]))/scale)
            dx += a[n+j+2*n]*3*(((x-xi[j,0])*(z-xi[j,2]))/scale)
            #y-derativate components numerator
            dy += a[n+j]*3*(((x-xi[j,0])*(y-xi[j,1]))/scale)
            dy += a[n+j+n]*3*((((y-xi[j,1])**2)/scale)+scale)
            dy += a[n+j+2*n]*3*(((y-xi[j,1])*(z-xi[j,2]))/scale)
            #z-derivative components numerator
            dz += a[n+j]*3*(((x-xi[j,0])*(z-xi[j,2]))/scale)
            dz += a[n+j+n]*3*(((z-xi[j,2])*(y-xi[j,1]))/scale)
            dz += a[n+j+2*n]*3*((((z-xi[j,2])**2)/scale)+scale)
            #x-derivative components denominator
            #fsg += a[n+j]*3*scale*(x-xi[j,0])
            #fsg += a[n+j+n]*3*scale*(y-xi[j,1])
            #fsg += a[n+j+2*n]*3*scale*(z-xi[j,2])
        #dx = -(1/np.cosh(fsg))**2*dx
        #dy = -(1/np.cosh(fsg))**2*dy
        #dz = -(1/np.cosh(fsg))**2*dz
        dxx = -dxx*(1/np.cosh(fsg))**2+2*dx**2*(1/np.cosh(fsg))**2*np.tanh(fsg)
        dxy = -dxy*(1/np.cosh(fsg))**2+2*dx*dy*(1/np.cosh(fsg))**2*np.tanh(fsg)
        dxz = -dxz*(1/np.cosh(fsg))**2+2*dx*dz*(1/np.cosh(fsg))**2*np.tanh(fsg)
        dyy = -dyy*(1/np.cosh(fsg))**2+2*dy**2*(1/np.cosh(fsg))**2*np.tanh(fsg)
        dyz = -dyz*(1/np.cosh(fsg))**2+2*dy*dz*(1/np.cosh(fsg))**2*np.tanh(fsg)
        dzz = -dzz*(1/np.cosh(fsg))**2+2*dz**2*(1/np.cosh(fsg))**2*np.tanh(fsg)
        fsg_dxx.append(dxx)
        fsg_dxy.append(dxy)
        fsg_dxz.append(dxz)
        fsg_dyy.append(dyy)
        fsg_dyz.append(dyz)
        fsg_dzz.append(dzz)
        #p0 = patch[0,:]
        #p1 = patch[-1,:]
        #d = ((p1[0]-p0[0])**2+(p1[1]-p0[1])**2+(p1[2]-p0[2])**2)**(1/2)
        r = ((x-patch_x[i])**2+(y-patch_y[i])**2+(z-patch_z[i])**2)**(1/2)
        #k = r/d
        #direc = np.array([x-p0[0],y-p0[1],z-p0[2]])/r
        value = (1-np.tanh(r))/(r+h)**q
        weights.append(value)
    weights = np.array(weights)
    total = np.sum(weights)
    res_xx = 0
    res_xy = 0
    res_xz = 0
    res_yy = 0
    res_yz = 0
    res_zz = 0
    for i in range(len(patch_x)):
        res_xx += (weights[i]/total)*fsg_dxx[i]
        res_xy += (weights[i]/total)*fsg_dxy[i]
        res_xz += (weights[i]/total)*fsg_dxz[i]
        res_yy += (weights[i]/total)*fsg_dyy[i]
        res_yz += (weights[i]/total)*fsg_dyz[i]
        res_zz += (weights[i]/total)*fsg_dzz[i]
    return np.array([[res_xx,res_xy,res_xz],[res_xy,res_yy,res_yz],[res_xz,res_yz,res_zz]])

def linesearch_d0(t,point=None,normal=None,patch_points=None,
                  b_coef=None,h=None,q=None,sol_mat=None,
                  patch_1=None,patch_2=None,patch_3=None):
    x  = point[0]
    y  = point[1]
    z  = point[2]
    nx = normal[0]
    ny = normal[1]
    nz = normal[2]
    xi = patch_points[:,:,0]
    yi = patch_points[:,:,1]
    zi = patch_points[:,:,2]
    xc = patch_1
    yc = patch_2
    zc = patch_3
    dx = (nx*t + x - xi)
    dy = (ny*t + y - yi)
    dz = (nz*t + z - zi)
    dxc = (nx*t + x - xc)
    dyc = (ny*t + y - yc)
    dzc = (nz*t + z - zc)
    r  = (dx**2 + dy**2 + dz**2)
    rc = (dxc**2 + dyc**2 + dzc**2)
    A  = sol_mat[:,:,0]
    B1 = sol_mat[:,:,1]
    B2 = sol_mat[:,:,2]
    B3 = sol_mat[:,:,3]
    C4 = b_coef[:,:,0]
    C1 = b_coef[:,:,1]
    C2 = b_coef[:,:,2]
    C3 = b_coef[:,:,3]
    finner = A*r**1.5 +\
             B1*(3.0*nx*t + 3.0*x - 3.0*xi)*r**0.5 +\
             B2*(3.0*ny*t + 3.0*y - 3.0*yi)*r**0.5 +\
             B3*(3.0*nz*t + 3.0*z - 3.0*zi)*r**0.5 +\
             C1*(nx*t + x) + C2*(ny*t + y) + C3*(nz*t + z) + C4
    finner = np.sum(finner,axis=1)
    f = -np.tanh(finner)
    w = (1 - np.tanh(rc**0.5))/(h + rc**0.5)**q
    f = (f*w)/np.sum(w)
    return np.sum(f)

def linesearch_d1(t,point=None,normal=None,patch_points=None,
                  b_coef=None,h=None,q=None,sol_mat=None,
                  patch_1=None,patch_2=None,patch_3=None):
    x  = point[0]
    y  = point[1]
    z  = point[2]
    nx = normal[0]
    ny = normal[1]
    nz = normal[2]
    xi = patch_points[:,:,0]
    yi = patch_points[:,:,1]
    zi = patch_points[:,:,2]
    xc = patch_1
    yc = patch_2
    zc = patch_3
    dx = (nx*t + x - xi)
    dy = (ny*t + y - yi)
    dz = (nz*t + z - zi)
    dxc = (nx*t + x - xc)
    dyc = (ny*t + y - yc)
    dzc = (nz*t + z - zc)
    r  = (dx**2 + dy**2 + dz**2)
    rc = (dxc**2 + dyc**2 + dzc**2)
    A  = sol_mat[:,:,0]
    B1 = sol_mat[:,:,1]
    B2 = sol_mat[:,:,2]
    B3 = sol_mat[:,:,3]
    C4 = b_coef[:,:,0]
    C1 = b_coef[:,:,1]
    C2 = b_coef[:,:,2]
    C3 = b_coef[:,:,3]
    finner = A*r**1.5 +\
             B1*(3.0*nx*t + 3.0*x - 3.0*xi)*r**0.5 +\
             B2*(3.0*ny*t + 3.0*y - 3.0*yi)*r**0.5 +\
             B3*(3.0*nz*t + 3.0*z - 3.0*zi)*r**0.5 +\
             C1*(nx*t + x) + C2*(ny*t + y) + C3*(nz*t + z) + C4
    finner = np.sum(finner,axis=1)
    f = -np.tanh(finner)
    w = (1 - np.tanh(rc**0.5))/(h + rc**0.5)**q

    wp = -q*(1 - np.tanh(rc**0.5))*(1.0*nx*dxc + 1.0*ny*dyc + 1.0*nz*dzc)/((h + rc**0.5)*(h + rc**0.5)**q*rc**0.5) -\
         (1 - np.tanh(rc**0.5)**2)*(1.0*nx*dxc + 1.0*ny*dyc + 1.0*nz*dzc)/((h + rc**0.5)**q*rc**0.5)

    fp = -(1 - np.tanh(finner)**2)*\
         np.sum(np.nan_to_num(A*(3.0*nx*dx + 3.0*ny*dy + 3.0*nz*dz)*r**0.5 +
                3.0*B1*nx*r**0.5 +
                B1*(3.0*nx*t + 3.0*x - 3.0*xi)*(1.0*nx*dx + 1.0*ny*dy + 1.0*nz*dz)/r**0.5 +
                3.0*B2*ny*r**0.5 +
                B2*(1.0*nx*dx + 1.0*ny*dy + 1.0*nz*dz)*(3.0*ny*t + 3.0*y - 3.0*yi)/r**0.5 +
                3.0*B3*nz*r**0.5 +
                B3*(1.0*nx*dx + 1.0*ny*dy + 1.0*nz*dz)*(3.0*nz*t + 3.0*z - 3.0*zi)/r**0.5 +
                C1*nx + C2*ny + C3*nz),axis=1)

    DFT = f*wp/np.sum(w) - f*w*np.sum(wp)/np.sum(w)**2 + w*fp/np.sum(w)
    DFT = np.sum(DFT)
    return DFT

def linesearch_d2(t,point=None,normal=None,patch_points=None,b_coef=None,h=None,q=None,sol_mat=None,patch_1=None,patch_2=None,patch_3=None):
    x  = point[0]
    y  = point[1]
    z  = point[2]
    nx = normal[0]
    ny = normal[1]
    nz = normal[2]
    xi = patch_points[:,:,0]
    yi = patch_points[:,:,1]
    zi = patch_points[:,:,2]
    xc = patch_1
    yc = patch_2
    zc = patch_3
    dx = (nx*t + x - xi)
    dy = (ny*t + y - yi)
    dz = (nz*t + z - zi)
    dxc = (nx*t + x - xc)
    dyc = (ny*t + y - yc)
    dzc = (nz*t + z - zc)
    r  = (dx**2 + dy**2 + dz**2)
    rc = (dxc**2 + dyc**2 + dzc**2)
    A  = sol_mat[:,:,0]
    B1 = sol_mat[:,:,1]
    B2 = sol_mat[:,:,2]
    B3 = sol_mat[:,:,3]
    C4 = b_coef[:,:,0]
    C1 = b_coef[:,:,1]
    C2 = b_coef[:,:,2]
    C3 = b_coef[:,:,3]
    finner = A*r**1.5 +\
             B1*(3.0*nx*t + 3.0*x - 3.0*xi)*r**0.5 +\
             B2*(3.0*ny*t + 3.0*y - 3.0*yi)*r**0.5 +\
             B3*(3.0*nz*t + 3.0*z - 3.0*zi)*r**0.5 +\
             C1*(nx*t + x) + C2*(ny*t + y) + C3*(nz*t + z) + C4
    finner = np.sum(np.nan_to_num(finner),axis=1)
    f = -np.tanh(finner)

    w = (1 - np.tanh(rc**0.5))/(h + rc**0.5)**q

    wp = -q*(1 - np.tanh(rc**0.5))*(1.0*nx*dxc + 1.0*ny*dyc + 1.0*nz*dzc)/((h + rc**0.5)*(h + rc**0.5)**q*rc**0.5) -\
         (1 - np.tanh(rc**0.5)**2)*(1.0*nx*dxc + 1.0*ny*dyc + 1.0*nz*dzc)/((h + rc**0.5)**q*rc**0.5)

    wpp = -(1.0*q*(np.tanh(rc**0.5) - 1)*(q*(nx*dxc + ny*dyc + nz*dzc)**2/((h + rc**0.5)*rc**1.0) -
           (nx**2 + ny**2 + nz**2)/rc**0.5 + (nx*dxc + ny*dyc + nz*dzc)**2/rc**1.5 +
           (nx*dxc + ny*dyc + nz*dzc)**2/((h + rc**0.5)*rc**1.0))/(h + rc**0.5) +
           2.0*q*(np.tanh(rc**0.5)**2 - 1)*(nx*dxc + ny*dyc + nz*dzc)**2/((h + rc**0.5)*rc**1.0) +
           (np.tanh(rc**0.5)**2 - 1)*(-1.0*(nx**2 + ny**2 + nz**2)/rc**0.5 +
           1.0*(nx*dxc + ny*dyc + nz*dzc)**2/rc**1.5 +
           2.0*(nx*dxc + ny*dyc + nz*dzc)**2*np.tanh(rc**0.5)/rc**1.0))/(h + rc**0.5)**q


    part = np.sum(np.nan_to_num(A*r**1.5 + 3.0*B1*dx*r**0.5 + 3.0*B2*dy*r**0.5 +
                  3.0*B3*dz*r**0.5 + C1*(nx*t + x) + C2*(ny*t + y) +
                  C3*(nz*t + z) + C4),axis=1)

    fp = -(1 - np.tanh(finner)**2)*\
         np.sum(np.nan_to_num(A*(3.0*nx*dx + 3.0*ny*dy + 3.0*nz*dz)*r**0.5 +
                3.0*B1*nx*r**0.5 +
                B1*(3.0*nx*t + 3.0*x - 3.0*xi)*(1.0*nx*dx + 1.0*ny*dy + 1.0*nz*dz)/r**0.5 +
                3.0*B2*ny*r**0.5 +
                B2*(1.0*nx*dx + 1.0*ny*dy + 1.0*nz*dz)*(3.0*ny*t + 3.0*y - 3.0*yi)/r**0.5 +
                3.0*B3*nz*r**0.5 +
                B3*(1.0*nx*dx + 1.0*ny*dy + 1.0*nz*dz)*(3.0*nz*t + 3.0*z - 3.0*zi)/r**0.5 +
                C1*nx + C2*ny + C3*nz),axis=1)

    fpp = (np.tanh(part)**2 - 1)*\
          (np.sum(np.nan_to_num(3.0*A*(nx**2 + ny**2 + nz**2)*r**0.5 +
                  3.0*A*(nx*dx + ny*dy + nz*dz)**2/r**0.5 +
                  6.0*B1*nx*(nx*dx + ny*dy + nz*dz)/r**0.5 +
                  3.0*B1*(nx**2 + ny**2 + nz**2)*dx/r**0.5 -
                  3.0*B1*dx*(nx*dx + ny*dy + nz*dz)**2/r**1.5 +
                  6.0*B2*ny*(nx*dx + ny*dy + nz*dz)/r**0.5 +
                  3.0*B2*(nx**2 + ny**2 + nz**2)*dy/r**0.5 -
                  3.0*B2*(nx*dx + ny*dy + nz*dz)**2*dy/r**1.5 +
                  6.0*B3*nz*(nx*dx + ny*dy + nz*dz)/r**0.5 +
                  3.0*B3*(nx**2 + ny**2 + nz**2)*dz/r**0.5 -
                  3.0*B3*(nx*dx + ny*dy + nz*dz)**2*dz/r**1.5 -
                  18.0*(1.0*A*(nx*dx + ny*dy + nz*dz)*r**0.5 +
                  B1*nx*r**0.5 +
                  1.0*B1*dx*(nx*dx + ny*dy + nz*dz)/r**0.5 +
                  B2*ny*r**0.5 +
                  1.0*B2*(nx*dx + ny*dy + nz*dz)*dy/r**0.5 +
                  B3*nz*r**0.5 +
                  1.0*B3*(nx*dx + ny*dy + nz*dz)*dz/r**0.5 +
                  0.333333333333333*C1*nx +
                  0.333333333333333*C2*ny +
                  0.333333333333333*C3*nz)**2),axis=1)*\
           np.tanh(part))
    DFT2 = (-(np.sum(wpp) - 2*np.sum(wp)**2/np.sum(w))*f*w/np.sum(w) + f*wpp - 2*f*np.sum(wp)*wp/np.sum(w) + w*fpp + 2*fp*wp - 2*w*fp*np.sum(wp)/np.sum(w))/np.sum(w)
    return np.sum(DFT2)



# smooth connections between trees in a network
from geomdl import BSpline
from geomdl import utilities
from geomdl.visualization import VisMPL
import numpy as np
from copy import deepcopy

def smooth_original(forest,curve_sample_size_min=5,curve_sample_size_max=21,curve_degree=3):
    #FOREST_CONNECT_COPY = deepcopy(forest)
    connections = []
    for network in range(forest.number_of_networks):
        if not np.any(forest.connections[network][0] is None):
            network_connections = []
            for tree in range(forest.trees_per_network[network]):
                tree_assignment = forest.assignments[network][tree]
                idx = max(forest.networks[network][tree].data[:,-1]) + 1
                #tree_connection_edges = np.zeros((len(forest.connections[network][0])*(curve_sample_size-1),forest.networks[network][tree].data.shape[1]))
                tree_connection_edges = []
                for edge in range(len(forest.connections[network][0])):
                    #print('Tree {}: Edge: {}'.format(tree,edge))
                    success = False
                    samples = curve_sample_size_min
                    # Check the angle formed
                    vec_1 = forest.networks[network][tree].data[tree_assignment[edge],0:3]-forest.networks[network][tree].data[tree_assignment[edge],3:6]
                    vec_2 = forest.connections[network][0][edge]-forest.networks[network][tree].data[tree_assignment[edge],3:6]
                    vec_1 = vec_1/np.linalg.norm(vec_1)
                    vec_2 = vec_2/np.linalg.norm(vec_2)
                    P0    = (forest.networks[network][tree].data[tree_assignment[edge],0:3]+forest.networks[network][tree].data[tree_assignment[edge],3:6])/2
                    P1    = forest.networks[network][tree].data[tree_assignment[edge],3:6]
                    ctr_0 = ((forest.networks[network][tree].data[tree_assignment[edge],0:3]*(3/4)+forest.networks[network][tree].data[tree_assignment[edge],3:6]*(1/4))).flatten().tolist()
                    ctr_1 = ((forest.networks[network][tree].data[tree_assignment[edge],0:3]+forest.networks[network][tree].data[tree_assignment[edge],3:6])/2).flatten().tolist()
                    if (np.arccos(np.dot(vec_1,vec_2))/np.pi)*180 > 120:
                        ctr_2 = forest.networks[network][tree].data[tree_assignment[edge],3:6].flatten().tolist()
                        ctr_3 = forest.connections[network][0][edge].tolist()
                    else:
                        A = 140
                        t = ((np.cos(np.pi*(A/180))-P0[0]*vec_2[0]-P0[1]*vec_2[1]-P0[2]*vec_2[2]+P1[0]*vec_2[0]+P1[1]*vec_2[1]+P1[2]*vec_2[2])/(vec_2[0]**2+vec_2[1]**2+vec_2[2]**2))
                        t = min(t,0.3)
                        conn_len = np.linalg.norm(forest.connections[network][0][edge]-forest.networks[network][tree].data[tree_assignment[edge],3:6])/2
                    #    vec_12_average = (vec_1+vec_2)/2
                        adj_0 = (P1-t*(conn_len)*vec_2)
                        if conn_len < forest.networks[network][tree].data[tree_assignment[edge],21]:
                            print("WARNING: Sharp Corner, lofting may fail")
                    #    vec_adj = adj_0 - P0/np.linalg.norm(adj_0 - P0)
                    #    vec_diff = vec_adj - vec_12_average
                    #    vec_diff = np.argwhere(vec_diff != 0).flatten()
                    #    vec_idx  = vec_diff[:2]
                    #    #AA = np.array([[vec_adj[vec_idx[0]],vec_12_average[vec_idx[0]]],
                    #    #               [vec_adj[vec_idx[1]],vec_12_average[vec_idx[1]]]])
                    #    #BB = np.array([[P0[vec_idx[0]]-P1[vec_idx[0]]],
                    #    #               [P0[vec_idx[1]]-P1[vec_idx[1]]]])
                    #    #AA_inv = np.linalg.inv(AA)
                    #    #sol = BB @ AA
                    #    sol  = (P0-P1)/(vec_adj - vec_12_average)
                    #    tt = sol[0]
                        ctr_2 = ((P0 + adj_0)/2).flatten().tolist()
                        ctr_3 = forest.connections[network][0][edge].tolist()
                    curve = BSpline.Curve()
                    curve.degree = curve_degree
                    #if (np.arccos(np.dot(vec_1,vec_2))/np.pi)*180 > 90:
                    curve.ctrlpts = [ctr_1,ctr_2,ctr_3]
                    #else:
                    #curve.ctrlpts = [ctr_0,ctr_1,ctr_2,ctr_3]
                    curve.knotvector = utilities.generate_knot_vector(curve.degree,len(curve.ctrlpts))
                    while samples <= curve_sample_size_max and not success:
                        curve.sample_size = samples
                        curve.evaluate()
                        pts = curve.evalpts
                        d0  = forest.networks[network][tree].data[tree_assignment[edge],12:15]
                        for p in range(len(pts)-1):
                            p0 = np.array(pts[p])
                            p1 = np.array(pts[p+1])
                            l  = np.linalg.norm(p1-p0)
                            d1 = (p1-p0)/l
                            A  = (abs(np.arccos(np.dot(-d0,d1)))/np.pi)*180
                            if A < 90:
                                success = False
                                samples += 2
                                break
                            else:
                                success = True
                            d0 = d1
                    if samples > curve_sample_size_max:
                        samples -= 2
                    for i in range(samples-1):
                        tmp = np.zeros(forest.networks[network][tree].data.shape[1])
                        tmp[0:3] = np.array(pts[i])
                        tmp[3:6] = np.array(pts[i+1])
                        tmp[21]  = forest.networks[network][tree].data[tree_assignment[edge],21]
                        tmp[20]  = np.linalg.norm(np.array(pts[i+1])-np.array(pts[i]))
                        tmp[12:15] = (np.array(pts[i+1])-np.array(pts[i]))/tmp[20]
                        tmp[26] = forest.networks[network][tree].data[tree_assignment[edge],26]+i+1
                        tmp[22] = forest.networks[network][tree].data[tree_assignment[edge],22]
                        tmp[-1] = idx
                        tmp[15] = -1
                        tmp[16] = -1
                        idx += 1
                        if i == 0:
                            tmp[17] = forest.networks[network][tree].data[tree_assignment[edge],-1]
                        elif i > 0:
                            tree_connection_edges[-1][15] = tmp[-1]
                            tree_connection_edges[-1][16] = -1
                            tmp[17] = tree_connection_edges[-1][-1]
                        #else:
                        #    tree_connection_edges[-1][15] = -1
                        #    tree_connection_edges[-1][16] = -1
                        #    tmp[17] = tree_connection_edges[-1][-1]
                        tree_connection_edges.append(tmp.tolist())
                network_connections.append(np.array(tree_connection_edges))
            connections.append(network_connections)
        else:
            connections.append([None])
    return connections, None

def smooth(forest,curve_sample_size_min=5,curve_sample_size_max=21,curve_degree=3):
    #FOREST_CONNECT_COPY = deepcopy(forest)
    connections = []
    splines     = []
    for network in range(forest.number_of_networks):
        if not np.any(forest.connections[network][0] is None):
            network_connections = []
            network_splines     = []
            for tree in range(forest.trees_per_network[network]):
                tree_assignment = forest.assignments[network][tree]
                idx = max(forest.networks[network][tree].data[:,-1]) + 1
                #tree_connection_edges = np.zeros((len(forest.connections[network][0])*(curve_sample_size-1),forest.networks[network][tree].data.shape[1]))
                tree_connection_edges = []
                for edge in range(len(forest.connections[network][0])):
                    #print('Tree {}: Edge: {}'.format(tree,edge))
                    success = False
                    samples = curve_sample_size_min
                    # Check the angle formed
                    vec_1 = forest.networks[network][tree].data[tree_assignment[edge],0:3]-forest.networks[network][tree].data[tree_assignment[edge],3:6]
                    vec_2 = forest.connections[network][0][edge]-forest.networks[network][tree].data[tree_assignment[edge],3:6]
                    vec_1 = vec_1/np.linalg.norm(vec_1)
                    vec_2 = vec_2/np.linalg.norm(vec_2)
                    print((np.arccos(np.dot(vec_1,vec_2))/np.pi)*180)
                    P0    = (forest.networks[network][tree].data[tree_assignment[edge],0:3]+forest.networks[network][tree].data[tree_assignment[edge],3:6])/2
                    P1    = forest.networks[network][tree].data[tree_assignment[edge],3:6]
                    ctr_0 = ((forest.networks[network][tree].data[tree_assignment[edge],0:3]*(3/4)+forest.networks[network][tree].data[tree_assignment[edge],3:6]*(1/4))).flatten().tolist()
                    #ctr_1 = ((forest.networks[network][tree].data[tree_assignment[edge],0:3]+forest.networks[network][tree].data[tree_assignment[edge],3:6])/2).flatten().tolist()
                    ctr_1 = P0.flatten().tolist()
                    if False:
                        #if (np.arccos(np.dot(vec_1,vec_2))/np.pi)*180 > 0:
                        ctr_2 = forest.networks[network][tree].data[tree_assignment[edge],3:6].flatten().tolist()
                        ctr_3 = forest.connections[network][0][edge].tolist()
                    else:
                        A = 120
                        t = ((np.cos(np.pi*(A/180))-P0[0]*vec_2[0]-P0[1]*vec_2[1]-P0[2]*vec_2[2]+P1[0]*vec_2[0]+P1[1]*vec_2[1]+P1[2]*vec_2[2])/(vec_2[0]**2+vec_2[1]**2+vec_2[2]**2))
                        t = min(t,0.3)
                        conn_len = np.linalg.norm(forest.connections[network][0][edge]-forest.networks[network][tree].data[tree_assignment[edge],3:6])/2
                    #    vec_12_average = (vec_1+vec_2)/2
                        adj_0 = (P1-t*(conn_len)*vec_2)
                        if conn_len < forest.networks[network][tree].data[tree_assignment[edge],21]:
                            print("WARNING: Sharp Corner, lofting may fail")
                    #    vec_adj = adj_0 - P0/np.linalg.norm(adj_0 - P0)
                    #    vec_diff = vec_adj - vec_12_average
                    #    vec_diff = np.argwhere(vec_diff != 0).flatten()
                    #    vec_idx  = vec_diff[:2]
                    #    #AA = np.array([[vec_adj[vec_idx[0]],vec_12_average[vec_idx[0]]],
                    #    #               [vec_adj[vec_idx[1]],vec_12_average[vec_idx[1]]]])
                    #    #BB = np.array([[P0[vec_idx[0]]-P1[vec_idx[0]]],
                    #    #               [P0[vec_idx[1]]-P1[vec_idx[1]]]])
                    #    #AA_inv = np.linalg.inv(AA)
                    #    #sol = BB @ AA
                    #    sol  = (P0-P1)/(vec_adj - vec_12_average)
                    #    tt = sol[0]
                        ctr_2 = ((P0+P1)/2).flatten().tolist()
                        ctr_3 = ((P1+adj_0)/2).flatten().tolist()
                        ctr_4 = forest.connections[network][0][edge].tolist()
                    curve = BSpline.Curve()
                    curve.degree = curve_degree
                    #if (np.arccos(np.dot(vec_1,vec_2))/np.pi)*180 > 90:
                    curve.ctrlpts = [ctr_1,ctr_2,ctr_3,ctr_4]
                    #else:
                    #curve.ctrlpts = [ctr_0,ctr_1,ctr_2,ctr_3]
                    curve.knotvector = utilities.generate_knot_vector(curve.degree,len(curve.ctrlpts))
                    network_splines.append(deepcopy(curve))
                    while samples <= curve_sample_size_max and not success:
                        curve.sample_size = samples
                        curve.evaluate()
                        pts = curve.evalpts
                        d0  = forest.networks[network][tree].data[tree_assignment[edge],12:15]
                        for p in range(len(pts)-1):
                            p0 = np.array(pts[p])
                            p1 = np.array(pts[p+1])
                            l  = np.linalg.norm(p1-p0)
                            d1 = (p1-p0)/l
                            A  = (abs(np.arccos(np.dot(-d0,d1)))/np.pi)*180
                            if A < 90:
                                success = False
                                samples += 2
                                break
                            else:
                                success = True
                            d0 = d1
                    if samples > curve_sample_size_max:
                        samples -= 2
                    for i in range(samples-1):
                        tmp = np.zeros(forest.networks[network][tree].data.shape[1])
                        tmp[0:3] = np.array(pts[i])
                        tmp[3:6] = np.array(pts[i+1])
                        tmp[21]  = forest.networks[network][tree].data[tree_assignment[edge],21]
                        tmp[20]  = np.linalg.norm(np.array(pts[i+1])-np.array(pts[i]))
                        tmp[12:15] = (np.array(pts[i+1])-np.array(pts[i]))/tmp[20]
                        tmp[26] = forest.networks[network][tree].data[tree_assignment[edge],26]+i+1
                        tmp[22] = forest.networks[network][tree].data[tree_assignment[edge],22]
                        tmp[-1] = idx
                        tmp[15] = -1
                        tmp[16] = -1
                        idx += 1
                        if i == 0:
                            tmp[17] = forest.networks[network][tree].data[tree_assignment[edge],-1]
                        elif i > 0:
                            tree_connection_edges[-1][15] = tmp[-1]
                            tree_connection_edges[-1][16] = -1
                            tmp[17] = tree_connection_edges[-1][-1]
                        #else:
                        #    tree_connection_edges[-1][15] = -1
                        #    tree_connection_edges[-1][16] = -1
                        #    tmp[17] = tree_connection_edges[-1][-1]
                        tree_connection_edges.append(tmp.tolist())

                network_connections.append(np.array(tree_connection_edges))
            connections.append(network_connections)
            splines.append(network_splines)
        else:
            connections.append([None])
    return connections, None, splines


def get_angle(V1,V2):
    return np.arccos(np.dot(V1,V2))*(180/np.pi)

def smooth_new(forest,max_angle=45):
    connections = []
    for network in range(forest.number_of_networks):
        if not np.any(forest.connections[network][0] is None):
            network_connections = []
            for tree in range(forest.trees_per_network[network]):
                idx = max(forest.networks[network][tree].data[:,-1]) + 1
                tree_connection_edges = []
                for edge in range(len(forest.connections[network][0])):
                    edge_1   = forest.assignments[network][0][edge]
                    edge_2   = forest.assignemnts[network][1][edge]
                    match_vector = forest.networks[network][0].data[edge_1,3:6] - forest.networks[network][1].data[edge_2,3:6]
                    match_vector = match_vector/np.linalg.norm(match_vector)
                    vector_1 = forest.networks[network][0].data[edge_1,12:15]
                    vector_2 = forest.networks[network][1].data[edge_2,12:15]
                    
    return connections, None



loft="""options = geometry.LoftNurbsOptions()

def bad_edges(model):
    # Determine the number of non-manifold and degenerate
    # edges within a sv solid model
    #
    # Parameters:
    #
    # model     (sv.modeling.POLYDATA): sv solid modeling object
    #
    # Returns
    #
    # bad_edges_number   (int): number of non-manifold & degenerate edges
    fe = vtk.vtkFeatureEdges()
    fe.FeatureEdgesOff()
    fe.BoundaryEdgesOn()
    fe.NonManifoldEdgesOn()
    fe.SetInputData(model.get_polydata())
    fe.Update()
    return fe.GetOutput().GetNumberOfCells()

def clean(model):
    # Merge duplicate points, clean up bad edges, perform basic filtering
    #
    # Parameters:
    #
    # model    (sv.modeling.POLYDATA): sv solid model
    #
    # Returns:
    #
    # model    (sv.modeling.POLYDATA): sv solid model with cleaned polydata
    #                                  surface mesh
    clean_filter = vtk.vtkCleanPolyData()
    clean_filter.PointMergingOn()
    clean_filter.SetInputData(model.get_polydata())
    clean_filter.Update()
    model.set_surface(clean_filter.GetOutput())
    return model

def tri(model):
    # Triangle Filter for sv solid model
    #
    # Parameters
    #
    # model    (sv.modeling.POLYDATA): sv solid model object
    #
    # Returns
    #
    # model    (sv.modeling.POLYDATA): sv solid model object with
    #                                  surface mesh containing only triangles
    tri_filter = vtk.vtkTriangleFilter()
    tri_filter.SetInputData(model.get_polydata())
    tri_filter.Update()
    model.set_surface(tri_filter.GetOutput())
    return model

def fill(model):
    # Fill holes within a sv solid model in order to close an open manifold
    #
    # Parameters
    #
    # model    (sv.modeling.POLYDATA): sv solid modeling object
    #
    # Returns
    #
    # model    (sv.modeling.POLYDATA): sv solid modeling object with a surface
    #                                  mesh containing no holes
    poly = vmtk.cap(surface=model.get_polydata(),use_center=False)
    model.set_surface(poly)
    return model

def surf_area(poly):
    # Calculate the area of a polydata surface
    #
    # Parameters
    #
    # poly    (vtk.PolyData): polydata object
    #
    # Returns
    #
    # surface_area   (float): area of the polydata surface mesh
    mass = vtk.vtkMassProperties()
    mass.SetInputData(poly)
    mass.Update()
    return mass.GetSurfaceArea()

def remesh(model,radius_factor=10):
    # Remesh the surface model using MMG.
    #
    # Parameters
    #
    # model  (sv.modeling.POLYDATA): SV solid modeling object
    # radius_factor           (int): scale mesh global minimum edge size
    #                                based on outlet radius
    #
    # Returns
    #
    # model  (sv.modeling.POLYDATA): remeshed model object
    face_ids = model.get_face_ids()
    smallest = 1e8
    biggest = 0
    for id in face_ids:
        if surf_area(model.get_face_polydata(id)) < smallest:
            smallest = surf_area(model.get_face_polydata(id))
        if surf_area(model.get_face_polydata(id)) > biggest:
            biggest = surf_area(model.get_face_polydata(id))
    radius = (smallest/3.14)**(1/2)
    hmin = radius/radius_factor
    hmax = radius
    print("Remeshing Model:\\nhmin: ----> {}\\nhmax ----> {}".format(hmin,hmax))
    remeshed_polydata = mesh_utils.remesh(model.get_polydata(),hmin=hmin,hmax=hmax)
    model.set_surface(remeshed_polydata)
    return model

def remesh_face(model,face_id,radius_scale=10):
    # Remesh Faces of a surface model using MMG.
    #
    # Parameters
    #
    # model  (sv.modeling.POLYDATA): SV solid modeling object
    # face_id                 (int): face_id index within solid model
    # radius_factor           (int): scale mesh global minimum edge size
    #                                based on outlet radius
    #
    # Returns
    #
    # model  (sv.modeling.POLYDATA): model with remeshed face
    face_poly = model.get_face_polydata(face_id)
    edge_size = (surf_area(face_poly)/3.14)**(1/2)/radius_scale
    print("Remeshing Face: {} ----> Edge Size: {}".format(face_id,edge_size))
    remeshed_poly = mesh_utils.remesh_faces(model.get_polydata(),[face_id],edge_size)
    model.set_surface(remeshed_poly)
    return model

def remesh_wall(model,wall_id):
    wall_poly = model.get_face_polydata(wall_id)
    cell_number = wall_poly.GetNumberOfCells()
    edge_size = ((surf_area(wall_poly)/cell_number)*0.5)**(1/2)
    remeshed_poly = mesh_utils.remesh_faces(model.get_polydata(),[wall_id],edge_size)
    model.set_surface(remeshed_poly)
    new_cell_number = model.get_face_polydata(wall_id).GetNumberOfCells()
    print("Face Elements: {} ----> Edge Size: {}".format(cell_number,new_cell_number))
    return model

def remesh_caps(model):
    # Remesh all caps of a surface model using MMG. By definition
    # this function will not remesh the walls of the surface model.
    #
    # Parameters
    #
    # model  (sv.modeling.POLYDATA): SV solid modeling object
    #
    # Returns
    #
    # model  (sv.modeling.POLYDATA): model with remeshed caps
    cap_ids = model.identify_caps()
    face_ids = model.get_face_ids()
    for i,c in enumerate(cap_ids):
        if c:
            model = remesh_face(model,face_ids[i])
    return model

def norm(model):
    # Determine the normal vectors along the
    # polydata surface.
    #
    # PARAMETERS
    # model    (sv.modeling.POLYDATA): SV solid modeling object
    #
    # Returns
    #
    # model    (sv.modeling.POLYDATA): SV solid model with calculated normals
    norm_filter = vtk.vtkPolyDataNormals()
    norm_filter.AutoOrientNormalsOn()
    norm_filter.ComputeCellNormalsOn()
    norm_filter.ConsistencyOn()
    norm_filter.SplittingOn()
    norm_filter.NonManifoldTraversalOn()
    norm_filter.SetInputData(model.get_polydata())
    norm_filter.Update()
    model.set_surface(norm_filter.GetOutput())
    return model

def loft(contours,num_pts=50,distance=False):
    # Generate an open lofted NURBS surface along a given
    # vessel contour group.
    #
    # PARAMETERS:
    # contours (list):  list of contour polydata objects defining one vessel.
    # num_pts  (int) :  number of sample points to take along each contour.
    # distance (bool):  flag to use distance based method for contour alignment
    #
    # Return
    #
    # loft_solid  (sv.modeling.POLYDATA): sv solid model for the open lofted surface
    for idx in range(len(contours)):
        contours[idx] = geometry.interpolate_closed_curve(polydata=contours[idx],number_of_points=num_pts)
        if idx != 0:
            contours[idx] = geometry.align_profile(contours[idx-1],contours[idx],distance)
    options = geometry.LoftNurbsOptions()
    loft_polydata = geometry.loft_nurbs(polydata_list=contours,loft_options=options)
    loft_solid = modeling.PolyData()
    loft_solid.set_surface(surface=loft_polydata)
    return loft_solid

def loft_all(contour_list):
    # Loft all vessels defining the total model that you want to create.
    #
    # PARAMETERS
    # contour_list: (list): list of lists that contain polydata contour groups
    #                      Example for two vessels:
    #
    #                      contour_list -> [[polydataContourObject1,polydataContourObject2],[polydataContourObject1,polydataContourObject2]]
    #
    # RETURNS:
    # lofts:        (list): list of open sv solid models of the lofted 3D surface. Note that
    #                       the loft is not yet capped.
    lofts = []
    for group in contour_list:
        contours,polydata =  clean_contours(group)
        lofts.append(loft(polydata))
    return lofts


def cap_all(loft_list):
    # Cap all lofted vessels.
    #
    # PARAMETERS:
    # loft_list  (list): list of sv modeling solid objects that are open lofts generated from
    #                    the 'loft_all' function.
    #
    # RETURNS:
    # capped     (list): list of capped solids
    capped = []
    for loft_solid in loft_list:
        capped_solid = modeling.PolyData()
        capped_solid.set_surface(vmtk.cap(surface=loft_solid.get_polydata(),use_center=False))
        capped_solid.compute_boundary_faces(angle=45)
        capped.append(capped_solid)
    return capped

def check_cap_solids(cap_solid_list):
    # Check capped solids for bad edges
    #
    # Parameters
    # cap_solid_list  (list): list of sv.modeling.POLYDATA solids
    #
    # Returns
    #
    # bad_edges_exist (bool): True/False value for if any bad edges
    #                         exist within any of the solids within
    #                         list
    for solid in cap_solid_list:
        if bad_edges(solid) > 0:
            return False
    return True

def clean_contours(contours):
    num = len(contours)-2
    new_contours = [contours[0]]
    new_poly     = [contours[0].get_polydata()]
    for i in range(1,len(contours)-1):
        n1 = np.array(new_contours[-1].get_normal())
        n2 = np.array(contours[i].get_normal())
        n3 = np.array(contours[i+1].get_normal())
        if ((np.arccos(np.dot(n1,n2.T))/np.pi)*180 > 10) or ((np.arccos(np.dot(n2,n3.T))/np.pi)*180 > 10) or check_connection(contours[i]):
            new_contours.append(contours[i])
            new_poly.append(contours[i].get_polydata())
    if len(new_contours) == 1:
        mid = len(contours)//2
        new_contours.append(contours[mid])
        new_poly.append(contours[mid].get_polydata())
    new_contours.append(contours[-1])
    new_poly.append(contours[-1].get_polydata())
    return new_contours, new_poly

def check_connection(contour,contour_list=contour_list):
    keep = False
    check_center = np.array(contour.get_center())
    for group in contour_list:
        c = np.array(group[0].get_center())
        if np.linalg.norm(check_center - c) < (group[0].get_radius()*2*np.pi+contour.get_radius()*2*np.pi):
            keep = True
            break
    return keep

def create_vessels(contour_list,attempts=5):
    # create seperate capped vessels for all contour groups defining a model of interest.
    #
    # PARAMETERS:
    # contour_list: (list): list of lists of contour polydata objects defining individual vessels
    #                       within the total model.
    # attemps:      (int) : the number of times that bad edges correction will be attemped during loft
    #                       alignment
    i = 0
    success = False
    while not success and i < attempts:
        lofts = loft_all(contour_list)
        cap_solids = cap_all(lofts)
        success = check_cap_solids(cap_solids)
        i += 1
    if success:
        print('Lofting Passed')
    else:
        print('Lofting Failed')
    return cap_solids

def show(model):
    polydata = model.get_polydata()

    colors = vtk.vtkNamedColors()
    background = 'white'
    mapper = vtk.vtkPolyDataMapper()
    actor  = vtk.vtkActor()
    mapper.SetInputDataObject(polydata)
    actor.SetMapper(mapper)
    renderer = vtk.vtkRenderer()
    renderer.SetBackground(colors.GetColor3d(background))

    render_window = vtk.vtkRenderWindow()
    render_window.AddRenderer(renderer)
    render_window.SetWindowName('Model View')

    interactor = vtk.vtkRenderWindowInteractor()
    interactor.SetRenderWindow(render_window)
    renderer.AddActor(actor)
    render_window.Render()
    interactor.Start()

capped_vessels = create_vessels(contour_list)
"""


old_loft="""options = geometry.LoftNurbsOptions()
{}
loft_surfaces = []
loft_solids = []
capped = []
capped_solids = []
faces = []
for contour_polydata_list in contour_polydata:
    loft_surfaces.append(geometry.loft_nurbs(polydata_list=contour_polydata_list,loft_options=options))
    loft_solids.append(modeling.PolyData())
    loft_solids[-1].set_surface(surface=loft_surfaces[-1])
    capped.append(vmtk.cap(surface=loft_solids[-1].get_polydata(),use_center=False))
    capped_solids.append(modeling.PolyData())
    capped_solids[-1].set_surface(surface=capped[-1])
    faces.append(capped_solids[-1].compute_boundary_faces(angle={}))
"""



import numpy as np
import numba as nb
from .utils import norm

@nb.jit(nopython=True)
def M(surface):
    N = surface.shape[0]
    D = surface.shape[1]
    M_matrix = np.zeros((N*4,N*4))
    for i in range(N):
        for j in range(i,N):
            M_matrix[i,j] = norm(surface[i,:] - surface[j,:])**3
            M_matrix[j,i] = norm(surface[i,:] - surface[j,:])**3

    for i in range(N):
        for j in range(N):
            for k in range(D):
               M_matrix[i,N+j+k*N] = 3*norm(surface[i,:] - surface[j,:])*\
                                           (surface[i,k] - surface[j,k])
               M_matrix[N+j+k*N,i] = 3*norm(surface[i,:] - surface[j,:])*\
                                           (surface[i,k] - surface[j,k])

    for i in range(N):
        for j in range(i,N):
            diff = surface[i,:] - surface[j,:]
            for k in range(D):
                for l in range(D):
                    if i==j:
                        M_matrix[N+j+l*N,N+i+k*N] = 0
                        M_matrix[N+i+k*N,N+j+l*N] = 0
                    elif k==l:
                        M_matrix[N+j+l*N,N+i+k*N] = -(3*(diff[l])**2/norm(surface[i,:] - surface[j,:]) +\
                                                    3*norm(surface[i,:] - surface[j,:]))
                        M_matrix[N+i+k*N,N+j+l*N] = -(3*(diff[l])**2/norm(surface[i,:] - surface[j,:]) +\
                                                    3*norm(surface[i,:] - surface[j,:]))
                    else:
                        M_matrix[N+j+l*N,N+i+k*N] = -3*diff[k]*diff[l]/norm(surface[i,:] - surface[j,:])
                        M_matrix[N+i+k*N,N+j+l*N] = -3*diff[k]*diff[l]/norm(surface[i,:] - surface[j,:])
    return M_matrix



import numpy as np
import numba as nb
import math

@nb.jit(nopython=True,cache=True,nogil=True)
def close(data,point,N=20):
    a = data[:, 0:3]
    b = data[:, 3:6]
    distances = np.sqrt(np.sum(np.square((a+b)/2-point),axis=1))
    if N >= len(distances):
        N = len(distances)
    close_edges = np.argsort(distances)[:N]
    distances=distances[close_edges]
    return close_edges, distances

#@nb.jit(nopython=True,cache=True,nogil=True)
def close_exact(data,point):
    line_direction = data[:,12:15]
    ss = np.array([np.dot(data[i,0:3] - point,line_direction[i,:]) for i in range(data.shape[0])])
    tt = np.array([np.dot(point - data[i,3:6],line_direction[i,:]) for i in range(data.shape[0])])
    decision = [[ss[i],tt[i],0] for i in range(len(ss))]
    hh = np.array([np.max(np.array(i)) for i in decision])
    cc = np.cross((point - data[:,0:3]),line_direction,axis=1)
    cd = np.linalg.norm(cc,axis=1)
    line_distances = np.hypot(hh,cd)
    vessel = np.argsort(line_distances)
    return vessel, line_distances
"""
def close_exact(data,point):
    line_direction = data[:,12:15]
    ss = np.array([np.dot(data[i,0:3] - point,line_direction[i,:]) for i in range(data.shape[0])])
    tt = np.array([np.dot(point - data[i,3:6],line_direction[i,:]) for i in range(data.shape[0])])
    decision = [[ss[i],tt[i],0] for i in range(len(ss))]
    hh = np.array([np.max(np.array(i)) for i in decision])
    cc = np.cross((point - data[:,0:3]),line_direction,axis=1)
    cd = np.linalg.norm(cc,axis=1)
    line_distances = np.hypot(hh,cd)
    vessel = np.argsort(line_distances)
    return vessel, line_distances
"""
@nb.jit(nopython=True,nogil=True)
def close_binary_vectorize(data,point):
    best_dist = np.inf
    best      = -1.0
    for vessel in data:
        line_direction = vessel[12:15].reshape(-1,1)
        diff = vessel[0:3]-point
        ss = np.dot(diff,line_direction)
        tt = np.dot(point-vessel[3:6],line_direction)
        hh = max([ss[0].item(),tt[0].item(),0])
        diff2 = (point-vessel[0:3]) #.reshape(-1,1)
        line_direction2 = line_direction.T
        cc = np.cross(diff2,line_direction2)
        cd = np.linalg.norm(cc)
        ld = math.hypot(hh,cd).item()
        if ld < best_dist:
            best_dist = ld
            best = vessel[-1]
    return best,best_dist

@nb.jit(nopython=True,nogil=True)
def close_binary_vectorize2(data,point):
    best_dist = np.inf
    best      = -1.0
    for vessel in data:
        diff = [0,0,0]
        diff2 = [0,0,0]
        diff3 = [0,0,0]
        cc = [0,0,0]
        line_direction = vessel[12:15]
        diff[0] = vessel[0]-point[0]
        diff[1] = vessel[1]-point[1]
        diff[2] = vessel[2]-point[2]
        ss = diff[0]*line_direction[0]+diff[1]*line_direction[1]+diff[2]*line_direction[2]
        #ss = np.dot(diff,line_direction)
        diff2[0] = point[0]-vessel[3]
        diff2[1] = point[1]-vessel[4]
        diff2[2] = point[2]-vessel[5]
        #tt = np.dot(point-vessel[3:6],line_direction)
        tt = diff2[0]*line_direction[0]+diff2[1]*line_direction[1]+diff2[2]*line_direction[2]
        #hh = max([ss,tt,0])
        if ss > tt and ss > 0:
            hh = ss
        elif tt > ss and tt>0:
            hh=tt
        else:
            hh=0.0
        #diff2 = (point-vessel[0:3]) #.reshape(-1,1)
        diff3[0] = point[0]-vessel[0]
        diff3[1] = point[1]-vessel[1]
        diff3[2] = point[2]-vessel[2]
        #line_direction2 = line_direction.T
        #cc = np.cross(diff2,line_direction2)
        cc[0] = diff3[1]*line_direction[2] - diff3[2]*line_direction[1]
        cc[1] = diff3[2]*line_direction[0] - diff3[0]*line_direction[2]
        cc[2] = diff3[0]*line_direction[1] - diff3[1]*line_direction[0]
        #cd = np.linalg.norm(cc)
        cd = (cc[0]**2+cc[1]*2+cc[2]**2)**(1/2)
        #ld = math.hypot(hh,cd).item()
        ld = (hh**2+cd**2)**(1/2)
        if ld < best_dist:
            best_dist = ld
            best = vessel[-1]
    return best,best_dist



import numpy as np
import numba as nb
from .m_matrix import M
from .n_matrix import N

@nb.jit(nopython=True)
def A(surface):
    n = surface.shape[0]
    d = surface.shape[1]
    M_matrix = M(surface)
    N_matrix = N(surface)
    A_matrix = np.zeros(((n+1)*4,(n+1)*4))
    A_matrix[:n*4,:n*4] = M_matrix
    A_matrix[:n*4,n*4:] = N_matrix
    A_matrix[n*4:,:n*4] = N_matrix.T
    A_inv = np.linalg.inv(A_matrix)
    M_inv = A_inv[:n*4,:n*4]
    N_inv = A_inv[:n*4,n*4:]
    K00 = np.zeros((n,n))
    K01 = np.zeros((n,n*d))
    K11 = np.zeros((n*d,n*d))
    K00[:,:] = M_inv[:n,:n]
    K01[:,:] = M_inv[:n,n:]
    K11[:,:] = M_inv[n:,n:]
    return A_inv,K00,K01,K11



import numba as nb
import numpy as np

@nb.jit(nopython=True,cache=True,nogil=True)
def basis(data, edges):
    data[edges, 12:15] = (data[edges, 3:6] -
                          data[edges, 0:3]) / np.linalg.norm(data[edges, 3:6] -
                                                             data[edges, 0:3])
    if data[edges, 14] == -1:
        data[edges, 6:9] = np.array([-1,0,0])
        data[edges, 9:12] = np.array([0,-1,0])
    else:
        data[edges, 6:9] = np.array([1-data[edges, 12]**2/(1+data[edges, 14]),
                                     (-data[edges, 12]*data[edges, 13])/(1+data[edges, 14]),
                                     -data[edges, 12]])
        data[edges, 9:12] = np.array([(-data[edges, 12]*data[edges, 13])/(1+data[edges, 14]),
                                      1 - data[edges, 13]**2/(1+data[edges, 14]),
                                      -data[edges, 13]])

def tangent_basis(normal,pt):
    normal = normal/np.linalg.norm(normal)
    p = np.random.random(3)
    r = (p - pt)/np.linalg.norm(p - pt)
    t1 = r - np.dot(r,normal)*normal
    t1 = t1/np.linalg.norm(t1)
    t2 = np.cross(t1.T,normal)
    return t1.reshape(1,-1),t2.reshape(1,-1),normal